# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import sys
from tempfile import NamedTemporaryFile
import unittest
from test import generic

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import configfile


class TestConfigFile(generic.TestCase):
    """
    Tests for the ConfigFile class in the configfile module
    """

    def test_save(self):
        """
        Saves the config file in the tmp directory
        """
        with NamedTemporaryFile() as cfgFile:
            cf = configfile.ConfigFile()
            self.assertTrue(cf.save(cfgFile.name))
            self.assertExists(cfgFile.name)

        #test fail routine in ConfigFile.save
        with NamedTemporaryFile() as fakeDir:
            self.assertFalse(cf.save(os.path.join(fakeDir.name, 'foo')))

    def test_load(self):
        """
        ConfigFile should be able to load its content from a previously
        saved ConfigFile object.
        """
        with NamedTemporaryFile() as cfgFile:
            original_cf = configfile.ConfigFile()
            key = "config_key"
            value = "config_value"
            original_cf.setStrValue(key, value)
            original_cf.save(cfgFile.name)

            cf = configfile.ConfigFile()
            cf.load(cfgFile.name)

            self.assertEqual(len(cf.keys()), len(original_cf.keys()))
            for k in original_cf.keys():
                with self.subTest(k = k):
                    #workaround for py.test3 2.5.1 doesn't support subTest
                    msg = 'k = %s' %k
                    self.assertTrue(cf.hasKey(k), msg)
                    self.assertEqual(original_cf.strValue(k), cf.strValue(k))

    def test_remapKey(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': '123',
                    'bar': '456'}
        #old key not in dict
        cfg.remapKey('notExistedKey', 'baz')
        self.assertEqual(cfg.strValue('foo'), '123')
        self.assertEqual(cfg.strValue('bar'), '456')

        #valid remap
        cfg.remapKey('foo', 'baz')
        self.assertEqual(cfg.strValue('foo'), '')
        self.assertEqual(cfg.strValue('baz'), '123')
        self.assertEqual(cfg.strValue('bar'), '456')

        #do not overwrite existing keys
        cfg.remapKey('baz', 'bar')
        self.assertEqual(cfg.strValue('baz'), '')
        self.assertEqual(cfg.strValue('bar'), '456')

    def test_remapKeyRegex(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'asdf.foo.qwertz': '123',
                    'foo.qwertz': '456',
                    'asdf.foo': '789',
                    'yxcv': 'jkl'}

        cfg.remapKeyRegex('foo', 'bar')
        self.assertEqual(cfg.strValue('asdf.bar.qwertz', 'WrongValue'), '123')
        self.assertEqual(cfg.strValue('bar.qwertz', 'WrongValue'), '456')
        self.assertEqual(cfg.strValue('asdf.bar', 'WrongValue'), '789')
        self.assertEqual(cfg.strValue('asdf.foo.qwertz', ''), '')
        self.assertEqual(cfg.strValue('foo.qwertz', ''), '')
        self.assertEqual(cfg.strValue('asdf.foo', ''), '')
        self.assertEqual(cfg.strValue('yxcv', 'WrongValue'), 'jkl')

    def test_hasKey(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': 'bar'}
        self.assertTrue(cfg.hasKey('foo'))
        self.assertFalse(cfg.hasKey('non_existend_key'))

    ############################################################################
    ###                               str_value                              ###
    ############################################################################

    def test_strValue(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': 'bar'}
        self.assertEqual(cfg.strValue('foo', 'default'), 'bar')

    def test_strValue_default(self):
        cfg = configfile.ConfigFile()
        self.assertEqual(cfg.strValue('non_existend_key', 'default'), 'default')

    def test_setStrValue(self):
        cfg = configfile.ConfigFile()
        cfg.setStrValue('foo', 'bar')
        self.assertDictEqual(cfg.dict, {'foo': 'bar'})

    ############################################################################
    ###                               int_value                              ###
    ############################################################################

    def test_intValue(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': '11'}
        self.assertEqual(cfg.intValue('foo', 22), 11)

    def test_intValue_default(self):
        cfg = configfile.ConfigFile()
        self.assertEqual(cfg.intValue('non_existend_key', 33), 33)

    def test_setIntValue(self):
        cfg = configfile.ConfigFile()
        cfg.setIntValue('foo', 44)
        self.assertDictEqual(cfg.dict, {'foo': '44'})

    ############################################################################
    ###                              bool_value                              ###
    ############################################################################

    def test_boolValue(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': 'true',
                    'bar': '1',
                    'baz': 'false',
                    'bla': '0'}
        self.assertEqual(cfg.boolValue('foo', False), True)
        self.assertEqual(cfg.boolValue('bar', False), True)
        self.assertEqual(cfg.boolValue('baz', True), False)
        self.assertEqual(cfg.boolValue('bla', True), False)

    def test_boolValue_default(self):
        cfg = configfile.ConfigFile()
        self.assertEqual(cfg.boolValue('non_existend_key', False), False)
        self.assertEqual(cfg.boolValue('non_existend_key', True), True)

    def test_setBoolValue(self):
        cfg = configfile.ConfigFile()
        cfg.setBoolValue('foo', True)
        cfg.setBoolValue('bar', False)
        self.assertDictEqual(cfg.dict, {'foo': 'true',
                                        'bar': 'false'})

    ############################################################################
    ###                           listValue                             ###
    ############################################################################

    def test_listValue_default(self):
        cfg = configfile.ConfigFile()
        self.assertListEqual(cfg.listValue('test', 'str:value', ['asdf']), ['asdf'])

    def test_listValue_int(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'aaa.size': '3',
                    'aaa.1.bla': '55',
                    'aaa.2.bla': '66',
                    'aaa.3.bla': '77'}
        self.assertListEqual(cfg.listValue('aaa', 'int:bla'), [55, 66, 77])

    def test_listValue_str(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'bbb.size': '3',
                    'bbb.1.value': 'foo',
                    'bbb.2.value': 'bar',
                    'bbb.3.value': 'baz'}
        self.assertListEqual(cfg.listValue('bbb', 'str:value'), ['foo', 'bar', 'baz'])

    def test_listValue_bool(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'ccc.size': '2',
                    'ccc.1.foo': 'true',
                    'ccc.2.foo': 'false'}
        self.assertListEqual(cfg.listValue('ccc', 'bool:foo'), [True, False])

    def test_listValue_tuple(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'ddd.size': '3',
                    'ddd.1.value': 'foo',
                    'ddd.1.type': '11',
                    'ddd.1.enabled': 'true',
                    'ddd.2.value': 'bar',
                    'ddd.2.type': '22',
                    'ddd.2.enabled': 'false',
                    'ddd.3.value': 'baz',
                    'ddd.3.type': '33',
                    'ddd.3.enabled': 'true'}
        self.assertListEqual(cfg.listValue('ddd', ('str:value', 'int:type', 'bool:enabled')),
                             [('foo', 11, True), ('bar', 22, False), ('baz', 33, True)])

    def test_listValue_tuple_missing_values(self):
        """
        Don't include missing values. Bug #521
        https://github.com/bit-team/backintime/issues/521
        """
        cfg = configfile.ConfigFile()
        cfg.dict = {'eee.size': '3',
                    'eee.1.value': 'foo',
                    'eee.1.enabled': 'true',
                    'eee.2.type': '22',
                    'eee.2.enabled': 'false',
                    'eee.3.value': 'baz',
                    'eee.3.type': '33'}
        self.assertListEqual(cfg.listValue('eee', ('str:value', 'int:type', 'bool:enabled')),
                             [('foo', 0, True), ('baz', 33, False)])

    def test_listValue_invalid_type(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'aaa.size': '3',
                    'aaa.1.value': '55',
                    'aaa.2.value': '66',
                    'aaa.3.value': '77'}
        with self.assertRaises(TypeError):
            cfg.listValue('aaa', 'non_existend_type:value')
        with self.assertRaises(TypeError):
            cfg.listValue('aaa', {'dict:value'})
        with self.assertRaises(TypeError):
            cfg.listValue('aaa', 1)

    def test_listValue_wrong_size(self):
        """
        Don't include empty values if size is wrong. Bug #521
        https://github.com/bit-team/backintime/issues/521
        """
        cfg = configfile.ConfigFile()
        cfg.dict = {'bbb.size': '4',
                    'bbb.1.value': 'foo',
                    'bbb.2.value': 'bar',
                    'bbb.3.value': 'baz'}
        self.assertListEqual(cfg.listValue('bbb', 'str:value'), ['foo', 'bar', 'baz'])

    def test_listValue_zero_count(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'ccc.size': '2',
                    'ccc.0.value': 'foo',
                    'ccc.1.value': 'bar',
                    'ccc.2.value': 'baz'}
        self.assertListEqual(cfg.listValue('ccc', 'str:value'), ['bar', 'baz'])

    def test_listValue_missing_values(self):
        """
        Don't include missing values. Bug #521
        https://github.com/bit-team/backintime/issues/521
        """
        cfg = configfile.ConfigFile()
        cfg.dict = {'ddd.size': '4',
                    'ddd.1.value': 'foo',
                    'ddd.2.value': 'bar',
                    'ddd.4.value': 'baz'}
        self.assertListEqual(cfg.listValue('ddd', 'str:value'),
                             ['foo', 'bar', 'baz'])

    def test_listValue_empty_list(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'eee.size': '0'}
        self.assertListEqual(cfg.listValue('eee', 'str:value',
                                           default = ['foo', 'bar']),
                             [])

    ############################################################################
    ###                           setListValue                             ###
    ############################################################################

    def test_setListValue_int(self):
        cfg = configfile.ConfigFile()
        cfg.setListValue('aaa', 'int:bla', [55, 66, 77])
        self.assertDictEqual(cfg.dict, {'aaa.size': '3',
                                        'aaa.1.bla': '55',
                                        'aaa.2.bla': '66',
                                        'aaa.3.bla': '77'})

    def test_setListValue_str(self):
        cfg = configfile.ConfigFile()
        cfg.setListValue('bbb', 'str:value', ['foo', 'bar', 'baz'])
        self.assertDictEqual(cfg.dict, {'bbb.size': '3',
                                        'bbb.1.value': 'foo',
                                        'bbb.2.value': 'bar',
                                        'bbb.3.value': 'baz'})

    def test_setListValue_bool(self):
        cfg = configfile.ConfigFile()
        cfg.setListValue('ccc', 'bool:foo', [True, False])
        self.assertDictEqual(cfg.dict, {'ccc.size': '2',
                                        'ccc.1.foo': 'true',
                                        'ccc.2.foo': 'false'})

    def test_setListValue_tuple(self):
        cfg = configfile.ConfigFile()
        cfg.setListValue('ddd', ('str:value', 'int:type', 'bool:enabled'),
                           [('foo', 11, True), ('bar', 22, False), ('baz', 33, True)])
        self.assertDictEqual(cfg.dict, {'ddd.size': '3',
                                        'ddd.1.value': 'foo',
                                        'ddd.1.type': '11',
                                        'ddd.1.enabled': 'true',
                                        'ddd.2.value': 'bar',
                                        'ddd.2.type': '22',
                                        'ddd.2.enabled': 'false',
                                        'ddd.3.value': 'baz',
                                        'ddd.3.type': '33',
                                        'ddd.3.enabled': 'true'})

    def test_setListValue_tuple_missing_values(self):
        cfg = configfile.ConfigFile()
        cfg.setListValue('ddd', ('str:value', 'int:type', 'bool:enabled'),
                           [('foo', 11, True), ('bar', 22), ('baz',)])
        self.assertDictEqual(cfg.dict, {'ddd.size': '3',
                                        'ddd.1.value': 'foo',
                                        'ddd.1.type': '11',
                                        'ddd.1.enabled': 'true',
                                        'ddd.2.value': 'bar',
                                        'ddd.2.type': '22',
                                        'ddd.3.value': 'baz'})

    def test_setListValue_remove_leftovers(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'eee.size': '5',
                    'eee.1.bla': '55',
                    'eee.2.bla': '66',
                    'eee.3.bla': '77',
                    'eee.4.bla': '88',
                    'eee.5.bla': '99'}
        cfg.setListValue('eee', 'int:bla', [55, 66, 77])
        self.assertDictEqual(cfg.dict, {'eee.size': '3',
                                        'eee.1.bla': '55',
                                        'eee.2.bla': '66',
                                        'eee.3.bla': '77'})

    def test_setListValue_remove_leftovers_tuple(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'fff.size': '5',
                    'fff.1.value': 'foo',
                    'fff.1.type': '11',
                    'fff.1.enabled': 'true',
                    'fff.2.value': 'bar',
                    'fff.2.type': '22',
                    'fff.2.enabled': 'false',
                    'fff.3.value': 'baz',
                    'fff.3.type': '33',
                    'fff.3.enabled': 'true',
                    'fff.4.value': 'boom',
                    'fff.4.type': '44',
                    'fff.4.enabled': 'true',
                    'fff.5.value': 'bam',
                    'fff.5.type': '55',
                    'fff.5.enabled': 'true'}
        cfg.setListValue('fff', ('str:value', 'int:type', 'bool:enabled'),
                           [('foo', 11, True), ('bar', 22), ('baz',)])
        self.assertDictEqual(cfg.dict, {'fff.size': '3',
                                        'fff.1.value': 'foo',
                                        'fff.1.type': '11',
                                        'fff.1.enabled': 'true',
                                        'fff.2.value': 'bar',
                                        'fff.2.type': '22',
                                        'fff.3.value': 'baz'})

    def test_setListValue_invalid_type_for_type_key(self):
        cfg = configfile.ConfigFile()
        with self.assertRaises(TypeError):
            cfg.setListValue('aaa', 'non_existend_type:value', ['foo',])
        with self.assertRaises(TypeError):
            cfg.setListValue('aaa', {'dict:value'}, ['foo',])
        with self.assertRaises(TypeError):
            cfg.setListValue('aaa', 1, ['foo',])

    def test_setListValue_invalid_type_for_value(self):
        cfg = configfile.ConfigFile()
        with self.assertRaises(TypeError):
            cfg.setListValue('bbb', 'str:value', 'foo')
        with self.assertRaises(TypeError):
            cfg.setListValue('bbb', 'str:value', {'foo': 'bar'})

    ############################################################################
    ###                            remove keys                               ###
    ############################################################################

    def test_remove_key(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': 'true',
                    'bar': '1',
                    'baz': 'false',
                    'bla': '0'}
        cfg.removeKey('bla')
        self.assertDictEqual(cfg.dict, {'foo': 'true',
                                        'bar': '1',
                                        'baz': 'false'})

    def test_remove_keys_start_with(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': 'true',
                    'bar': '1',
                    'baz': 'false',
                    'bla': '0'}
        cfg.removeKeysStartsWith('ba')
        self.assertDictEqual(cfg.dict, {'foo': 'true',
                                        'bla': '0'})

    def test_remove_keys_start_with_not_matching_prefix(self):
        cfg = configfile.ConfigFile()
        cfg.dict = {'foo': 'true',
                    'bar': '1',
                    'baz': 'false',
                    'bla': '0'}
        cfg.removeKeysStartsWith('not_matching')
        self.assertDictEqual(cfg.dict, {'foo': 'true',
                                        'bar': '1',
                                        'baz': 'false',
                                        'bla': '0'})


class TestConfigFileWithProfiles(generic.TestCase):
    def setUp(self):
        super(TestConfigFileWithProfiles, self).setUp()
        self.cfg = configfile.ConfigFileWithProfiles('DefaultProfileName')
        self.cfg.addProfile('foo')
        self.cfg.addProfile('bar')
        self.cfg.addProfile('baz')

    def test_load(self):
        """
        ConfigFile should be able to load its content from a previously
        saved ConfigFile object.
        """
        with NamedTemporaryFile() as cfgFile:
            origCfg = configfile.ConfigFileWithProfiles()
            origCfg.setIntValue('profiles.version', 1)
            key = "config_key"
            value = "config_value"
            origCfg.setProfileStrValue(key, value)
            origCfg.setProfileStrValue(key, value, profile_id = '2')
            origCfg.save(cfgFile.name)

            self.cfg.load(cfgFile.name)

            self.assertEqual(len(self.cfg.keys()), len(origCfg.keys()))
            for k in origCfg.keys():
                with self.subTest(k = k):
                    #workaround for py.test3 2.5.1 doesn't support subTest
                    msg = 'k = %s' %k
                    self.assertTrue(self.cfg.hasKey(k), msg)
                    self.assertEqual(origCfg.strValue(k), self.cfg.strValue(k))

    def test_profiles(self):
        emptyCfg = configfile.ConfigFileWithProfiles()
        self.assertListEqual(emptyCfg.profiles(), ['1',])

        self.assertListEqual(self.cfg.profiles(), ['1', '2', '3', '4'])

        self.cfg.removeProfile('3')
        self.assertListEqual(self.cfg.profiles(), ['1', '2', '4'])

    def test_profilesSortedByName(self):
        self.assertListEqual(self.cfg.profilesSortedByName(), ['3', '4', '1', '2'])

    def test_current_profile(self):
        self.assertEqual(self.cfg.currentProfile(), '1')

        self.assertTrue(self.cfg.setCurrentProfile(4))
        self.assertEqual(self.cfg.currentProfile(), '4')

        self.assertTrue(self.cfg.setCurrentProfile('3'))
        self.assertEqual(self.cfg.currentProfile(), '3')

        self.assertFalse(self.cfg.setCurrentProfile('9'))
        self.assertEqual(self.cfg.currentProfile(), '3')

    def test_current_profile_by_name(self):
        self.assertEqual(self.cfg.currentProfile(), '1')

        self.assertTrue(self.cfg.setCurrentProfileByName('bar'))
        self.assertEqual(self.cfg.currentProfile(), '3')

        self.assertFalse(self.cfg.setCurrentProfileByName('NotExistingProfile'))
        self.assertEqual(self.cfg.currentProfile(), '3')

    def test_profileExists(self):
        self.assertTrue(self.cfg.profileExists('2'))
        self.assertTrue(self.cfg.profileExists(3))
        self.assertFalse(self.cfg.profileExists('9'))
        self.assertFalse(self.cfg.profileExists(10))

    def test_profileExistsByName(self):
        self.assertTrue(self.cfg.profileExistsByName('foo'))
        self.assertFalse(self.cfg.profileExistsByName('NotExistingProfile'))

    def test_profileName(self):
        self.assertEqual(self.cfg.profileName('1'), 'DefaultProfileName')
        self.assertEqual(self.cfg.profileName('2'), 'foo')
        self.assertEqual(self.cfg.profileName(3), 'bar')

        self.assertEqual(self.cfg.profileName(), 'DefaultProfileName')
        self.cfg.setCurrentProfile('3')
        self.assertEqual(self.cfg.profileName(), 'bar')

        self.assertEqual(self.cfg.profileName('4'), 'baz')
        del self.cfg.dict['profile4.name']
        self.assertEqual(self.cfg.profileName('4'), 'Profile 4')

    def test_addProfile(self):
        #add already existing profile
        self.assertIsNone(self.cfg.addProfile('foo'))

        #new valid profile
        self.assertEqual(self.cfg.addProfile('asdf'), '5')

        #new valid profile fill an old profile ID
        self.cfg.removeProfile('3')
        self.assertEqual(self.cfg.addProfile('qwertz'), '3')

    def test_removeProfile(self):
        for profile in self.cfg.profiles():
            self.cfg.setProfileStrValue('foo', 'bar', profile)

        self.assertFalse(self.cfg.removeProfile('9'))

        self.assertTrue(self.cfg.removeProfile('3'))
        self.assertNotIn('3', self.cfg.profiles())
        self.assertNotIn('profile3.foo', self.cfg.dict)

        self.cfg.setCurrentProfile('4')
        self.assertTrue(self.cfg.removeProfile())
        self.assertNotIn('4', self.cfg.profiles())
        self.assertNotIn('profile4.foo', self.cfg.dict)
        self.assertEqual(self.cfg.currentProfile(), '1')

        self.assertTrue(self.cfg.removeProfile(2))
        self.assertNotIn('2', self.cfg.profiles())
        self.assertNotIn('profile2.foo', self.cfg.dict)

        self.assertFalse(self.cfg.removeProfile())

    def test_setProfileName(self):
        self.assertFalse(self.cfg.setProfileName('foo', '3'))
        self.assertEqual(self.cfg.profileName('3'), 'bar')

        self.assertTrue(self.cfg.setProfileName('newName', '4'))
        self.assertEqual(self.cfg.profileName('4'), 'newName')

        self.assertTrue(self.cfg.setProfileName('otherName', 3))
        self.assertEqual(self.cfg.profileName('3'), 'otherName')

        self.assertTrue(self.cfg.setProfileName('thirdName'))
        self.assertEqual(self.cfg.profileName('1'), 'thirdName')

    def test_get_profile_key(self):
        self.assertEqual(self.cfg.profileKey('foo'), 'profile1.foo')
        self.assertEqual(self.cfg.profileKey('foo', '2'), 'profile2.foo')
        self.assertEqual(self.cfg.profileKey('foo', 3), 'profile3.foo')

    def test_removeProfileKey(self):
        for profile in self.cfg.profiles():
            self.cfg.setProfileStrValue('foo', 'bar', profile)

        self.assertIn('profile1.foo', self.cfg.dict)
        self.cfg.removeProfileKey('foo')
        self.assertNotIn('profile1.foo', self.cfg.dict)

        self.assertIn('profile3.foo', self.cfg.dict)
        self.cfg.removeProfileKey('foo', '3')
        self.assertNotIn('profile3.foo', self.cfg.dict)

    def test_removeProfileKeysStartsWith(self):
        for profile in self.cfg.profiles():
            self.cfg.setProfileStrValue('foo', 'bar', profile)

        self.assertIn('profile1.foo', self.cfg.dict)
        self.cfg.removeProfileKeysStartsWith('f')
        self.assertNotIn('profile1.foo', self.cfg.dict)

        self.assertIn('profile3.foo', self.cfg.dict)
        self.cfg.removeProfileKeysStartsWith('f', '3')
        self.assertNotIn('profile3.foo', self.cfg.dict)

    def test_remapProfileKey(self):
        for profile in self.cfg.profiles():
            self.cfg.setProfileStrValue('foo', '123', profile)
            self.cfg.setProfileStrValue('bar', '456', profile)

        # old key not in cfg
        self.cfg.remapProfileKey('notExistedKey', 'baz', 1)
        self.assertEqual(self.cfg.profileStrValue('foo', '', 1), '123')
        self.assertEqual(self.cfg.profileStrValue('bar', '', 1), '456')


        # valid remap
        self.cfg.remapProfileKey('foo', 'baz', 1)
        self.assertEqual(self.cfg.profileStrValue('foo', '', 1), '')
        self.assertEqual(self.cfg.profileStrValue('baz', '', 1), '123')
        self.assertEqual(self.cfg.profileStrValue('foo', '', 2), '123')
        self.assertEqual(self.cfg.profileStrValue('foo', '', 3), '123')

        # do not overwrite existing keys
        self.cfg.remapProfileKey('foo', 'bar', 1)
        self.assertEqual(self.cfg.profileStrValue('foo', '', 2), '123')
        self.assertEqual(self.cfg.profileStrValue('baz', '', 2), '')
        self.assertEqual(self.cfg.profileStrValue('bar', '', 2), '456')

    def test_hasProfileKey(self):
        for profile in self.cfg.profiles():
            self.cfg.setProfileStrValue('foo', 'bar', profile)

        self.assertTrue(self.cfg.hasProfileKey('foo'))
        self.assertFalse(self.cfg.hasProfileKey('baz'))

        self.assertTrue(self.cfg.hasProfileKey('foo', '3'))
        self.assertFalse(self.cfg.hasProfileKey('baz', '3'))

    def test_set_profile_value(self):
        methods =  {'str':  ('foo', 'FOO'),
                    'int':  ('bar', 123),
                    'bool': ('baz', True)}

        for profile in (None, '3'):
            for t in methods:
                with self.subTest(profile = profile, t = t):
                    #workaround for py.test3 2.5.1 doesn't support subTest
                    msg = 'profile = {}, t = {}'.format(profile, t)
                    key, value = methods[t]
                    setFunc = getattr(self.cfg, 'setProfile{}Value'.format(t.capitalize()))
                    getFunc = getattr(self.cfg, 'profile{}Value'.format(t.capitalize()))
                    setFunc(key, value, profile_id = profile)
                    self.assertEqual(getFunc(key, profile_id = profile), value, msg)
            with self.subTest(profile = profile):
                #workaround for py.test3 2.5.1 doesn't support subTest
                msg = 'profile = {}'.format(profile)
                self.cfg.setProfileListValue('bla',
                                                'str:value',
                                                ['ASDF', 'QWERTZ'],
                                                profile_id = profile)
                result = self.cfg.profileListValue('bla',
                                                         'str:value',
                                                         profile_id = profile)
                self.assertListEqual(result, ['ASDF', 'QWERTZ'], msg)


if __name__ == '__main__':
    unittest.main()



# SPDX-FileCopyrightText: © 2023 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Tests using several linters.

Linter tests are skipped on machines where the linters are not available. As an
exception only on TravisCI tests are forced to run and will fail if linters not
available.
"""
import unittest
import os
import pathlib
import subprocess
import shutil
from typing import Iterable
from packaging import version

BASE_REASON = ('Using package {0} is mandatory on TravisCI, on '
               'other systems it runs only if `{0}` is available.')

ON_TRAVIS = os.environ.get('TRAVIS', '') == 'true'
TRAVIS_REASON = ('Running linter tests is mandatory on TravisCI only. On '
                 'other machines they will run only if linters available.')

PYLINT_AVAILABLE = shutil.which('pylint') is not None
RUFF_AVAILABLE = shutil.which('ruff') is not None
FLAKE8_AVAILABLE = shutil.which('flake8') is not None
REUSE_AVAILABLE = shutil.which('reuse') is not None

ANY_LINTER_AVAILABLE = any((
    PYLINT_AVAILABLE,
    RUFF_AVAILABLE,
    FLAKE8_AVAILABLE,
))

# "common" directory
_base_dir = pathlib.Path(__file__).resolve().parent.parent

# Files in this lists will get the full battery of linters and rule sets.
full_test_files = [_base_dir / fp for fp in (
    'bitbase.py',
    'languages.py',
    'schedule.py',
    'singleton.py',
    'ssh_max_arg.py',
    'version.py',
    'test/test_lint.py',
    'test/test_mount.py',
    'test/test_singleton.py',
    'test/test_uniquenessset.py',
)]

# Not all linters do respect PEP8 (e.g. ruff, PyLint)
PEP8_MAX_LINE_LENGTH = 79


def create_pylint_cmd(include_error_codes=None):
    """Create the pylint base command for later use with subprocess.run().

    Args:
        include_error_codes: List of exclusive rules to check for. If empty all
            default rules are checked.
    """

    cmd = [
        'pylint',
        # Make sure BIT modules can be imported (to detect "no-member")
        '--init-hook=import sys;'
        'sys.path.insert(0, "./../qt");'
        'sys.path.insert(0, "./../common");',
        # Storing results in a pickle file is unnecessary
        '--persistent=n',
        # autodetec number of parallel jobs
        '--jobs=0',
        # Disable scoring  ("Your code has been rated at xx/10")
        '--score=n',
        # prevent false-positive no-module-member errors
        '--extension-pkg-allow-list=PyQt6,PyQt6.QtCore',
        # Because of globally installed GNU gettext functions
        '--additional-builtins=_,ngettext',
        # PEP8 conform line length (see PyLint Issue #3078)
        f'--max-line-length={PEP8_MAX_LINE_LENGTH}',
        # Whitelist variable names
        '--good-names=idx,fp',
        # '--reports=yes',
    ]

    if include_error_codes:
        # Deactivate all checks by default
        cmd.append('--disable=all')
        # Include specific codes only
        cmd.append('--enable=' + ','.join(include_error_codes))

    return cmd


@unittest.skipUnless(ON_TRAVIS or ANY_LINTER_AVAILABLE, TRAVIS_REASON)
class MirrorMirrorOnTheWall(unittest.TestCase):
    """Check all py-files in the package (incl. test files) for lints,
    potential bugs and if they are compliant to the coding styles (e.g. PEP8).
    """

    @classmethod
    def _collect_py_files(cls) -> Iterable[pathlib.Path]:
        """All py-files related to that distribution package.

        Dev note (2023-11): Use package metadata after migration to
        pyproject.toml.
        """
        path = pathlib.Path.cwd()

        # Make sure we are inside the test folder
        if path.name in ['qt', 'common']:  # happens e.g. on TravisCI
            path = path / 'test'

        if not path.name.startswith('test'):
            raise RuntimeError('Something went wrong. The test should run '
                               'inside the test folder but the current folder '
                               f'is {path}.')

        # Workaround
        path = path.parent

        # Find recursive all py-files.
        py_files = path.rglob('*.py')

        # Exclude full test files
        return filter(lambda fp: fp not in full_test_files, py_files)

    @classmethod
    def setUpClass(cls):
        cls.collected_py_files = cls._collect_py_files()

    def test005_ensure_linter_versions(self):
        """Workaround to ensure the correct linter versions are used.

        For sure there are better ways to solve this. But migration to a
        standard python package format need to be done first. See #1575.
        Until then this test will spare some hours of work, e.g. fixing linter
        errors (from out-dated linters) that are not relevant anymore in
        modern lintern versions.

        Another location where linter versions are relevant is CONTRIBUTING.md.
        """

        if PYLINT_AVAILABLE:
            version_target = version.parse('3.3.0')

            proc = subprocess.run(
                ['pylint', '--version'],
                capture_output=True,
                text=True,
                check=True)

            version_string = proc.stdout.split('\n')[0].replace('pylint ', '')
            version_actual = version.parse(version_string)

            self.assertTrue(
                version_actual >= version_target,
                f'PyLint version is {version_actual} but need to '
                f'be {version_target} or higher.')

        if RUFF_AVAILABLE:
            version_target = version.parse('0.6.0')

            proc = subprocess.run(
                ['ruff', '--version'],
                capture_output=True,
                text=True,
                check=True)

            version_string = proc.stdout.split('\n')[0].replace('ruff ', '')
            version_actual = version.parse(version_string)

            self.assertTrue(
                version_actual >= version_target,
                f'Ruff version is {version_actual} but need to '
                f'be {version_target} or higher.')

    @unittest.skipUnless(RUFF_AVAILABLE, BASE_REASON.format('ruff'))
    def test010_ruff_default_ruleset(self):
        """Ruff in default mode."""

        # ATTENTIION: Some settings are found in pyproject.toml
        cmd = [
            'ruff',
            'check',
            # Additionally activate subset of special rules:
            # - PyLint (PL)
            # - PyCodestyle (E, W)
            # - flake8-gettext (INT)
            # - useless noqua (RUF100)
            '--extend-select=PL,E,W,INT,RUF100',
            # Ignore: redefined-loop-name
            '--ignore=PLW2901',
            '--line-length', str(PEP8_MAX_LINE_LENGTH),
            # Because of globally installed GNU gettext functions
            '--config', 'builtins=["_", "ngettext"]',
            # Ruff counting branches different from PyLint.
            # See: <https://www.reddit.com/r/learnpython/comments/
            #      1buojae/comment/kxu0mp3>
            '--config', 'pylint.max-branches=13',
            '--config', 'flake8-quotes.inline-quotes = "single"',
            # one error per line (no context lines)
            '--output-format=concise',
            '--quiet',
        ]

        cmd.extend(full_test_files)

        proc = subprocess.run(
            cmd,
            check=False,
            universal_newlines=True,
            capture_output=True
        )

        # No errors other then linter rules
        self.assertIn(proc.returncode, [0, 1], proc.stderr)

        error_n = len(proc.stdout.splitlines())
        if error_n > 0:
            print(proc.stdout)

        self.assertEqual(0, error_n, f'Ruff found {error_n} problem(s).')

        # any other errors?
        self.assertEqual(proc.stderr, '')

    @unittest.skipUnless(FLAKE8_AVAILABLE, BASE_REASON.format('flake8'))
    def test020_flake8_default_ruleset(self):
        """Flake8 in default mode."""
        cmd = [
            'flake8',
            f'--max-line-length={PEP8_MAX_LINE_LENGTH}',
            '--builtins=_,ngettext',
            # '--enable-extensions='
        ]

        cmd.extend(full_test_files)

        proc = subprocess.run(
            cmd,
            check=False,
            universal_newlines=True,
            capture_output=True
        )

        error_n = len(proc.stdout.splitlines())
        if error_n > 0:
            print(proc.stdout)

        self.assertEqual(0, error_n, f'Flake8 found {error_n} problem(s).')

        # any other errors?
        self.assertEqual(proc.stderr, '')

    @unittest.skipUnless(PYLINT_AVAILABLE, BASE_REASON.format('PyLint'))
    def test030_pylint_default_ruleset(self):
        """Use Pylint with all default rules to check specific files.
        """

        cmd = create_pylint_cmd()

        # Add py-files
        cmd.extend(full_test_files)

        r = subprocess.run(
            cmd,
            check=False,
            universal_newlines=True,
            capture_output=True)

        # Count lines except module headings
        error_n = len(list(filter(lambda line: not line.startswith('*****'),
                                  r.stdout.splitlines())))
        print(r.stdout)

        self.assertEqual(0, error_n, f'PyLint found {error_n} problems.')

        # any other errors?
        self.assertEqual(r.stderr, '')

    @unittest.skipUnless(PYLINT_AVAILABLE, BASE_REASON.format('PyLint'))
    def test050_pylint_exclusive_ruleset(self):
        """Use Pylint to check for specific rules only.

        Some facts about PyLint
         - It is one of the slowest available linters.
         - It is able to catch lints other linters miss.
        """

        # Explicit activate checks
        err_codes = [
            'C0305',  # trailing-newlines
            'C0325',  # superfluous-parens
            'C0410',  # multiple-imports
            'C0303',  # trailing-whitespace
            'E0100',  # init-is-generator
            'E0101',  # return-in-init
            'E0102',  # function-redefined
            'E0103',  # not-in-loop
            'E0106',  # return-arg-in-generator
            'E0213',  # no-self-argument
            'E0401',  # import-error
            'E0602',  # undefined-variable
            'E1101',  # no-member
            'I0021',  # useless-suppression
            'W0123',  # eval-used
            'W0237',  # arguments-renamed
            'W0311',  # bad-indentation
            'W0404',  # reimported
            'W0611',  # unused-import
            'W0612',  # unused-variable
            'W0614',  # unused-wildcard-import
            'W0707',  # raise-missing-from
            'W1301',  # unused-format-string-key
            'W1401',  # anomalous-backslash-in-string (invalid escape sequence)
            'W1515',  # forgotten-debug-statement
            'W4902',  # deprecated-method
            'W4904',  # deprecated-class
            'R0202',  # no-classmethod-decorator
            'R0203',  # no-staticmethod-decorator
            # See PyLint bugs:
            # https://github.com/pylint-dev/pylint/issues/214
            # https://github.com/pylint-dev/pylint/issues/7920
            # 'R0801',  # duplicate-code

            # Enable asap. This list is a selection of existing (not all!)
            # problems currently existing in the BIT code base. Quite easy to
            # fix because their count is low.
            # 'W0237',  # arguments-renamed
            # 'W0221',  # arguments-differ
            # 'W0603',  # global-statement
        ]

        cmd = create_pylint_cmd(err_codes)

        # Add py-files
        cmd.extend(self._collect_py_files())

        r = subprocess.run(
            cmd,
            check=False,
            universal_newlines=True,
            capture_output=True)

        # Count lines except module headings and output about duplicate code
        error_n = len(list(filter(
            lambda line: line[:2] not in ('**', '  ', '==', ' (', ''),
            r.stdout.splitlines())))
        print(r.stdout)

        self.assertEqual(0, error_n, f'PyLint found {error_n} problems.')

        # any other errors?
        self.assertEqual(r.stderr, '')

    @unittest.skipUnless(REUSE_AVAILABLE, BASE_REASON.format('REUSE'))
    def test060_reuse(self):
        """The reuse linter check license and copyright information in the
        repository.

        The info need to be complete and available for all files. The
        info need to be provided as meta data conforming the SPDX standard.
        """
        proc = subprocess.run(
            ['reuse', 'lint', '--lines'],
            check=False,
            universal_newlines=True,
            capture_output=True
        )

        error_n = len(proc.stdout.splitlines())
        if error_n > 0:
            print(proc.stdout)

        self.assertEqual(
            0, error_n, f'REUSE linter found {error_n} problem(s).')

        # any other errors?
        self.assertEqual(proc.stderr, '')



# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Centralize management about the version.

That file is a workaround until the project migrated to a Python build-system.
See Issue #1575 for details about that migration.
"""
import re

# Version string regularyly used by the application and presented to users.
__version__ = '1.5.4'


def is_release_candidate() -> bool:
    """Test if the current version is a release candidate.

    It is the case if the version string ends with lower case ``rc`` and
    optionally with a number.
    """
    return bool(re.search(r'^.+rc\d+$', __version__))



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import pluginmanager
import logger
import gettext
from subprocess import Popen, PIPE
from exceptions import StopException

_ = gettext.gettext


class UserCallbackPlugin(pluginmanager.Plugin):
    """ Executes a script file at different backup steps to customize behavior

    Back In Time allows to inform plugins (implemented in Python
    files) about different steps ("events") in the backup process
    via the :py:class:`pluginmanager.PluginManager`.

    This plugin calls a user-defined script file ("user-callback"). By default
    that file is located in the config folder `$XDG_CONFIG_HOME/backintime/`
    or another folder if command line optioni `--config` is used.

    The user-callback script is called with up to five positional arguments:

    1. The profile ID from the config file (1=Main Profile, ...)
    2. The profile name (from the config file)
    3. A numeric code to indicate the reason why Back In Time
       calls the script (see the method implementation
       for details about the numeric code)
    4. Error code (only if argument 3 has the value "4")
       or snapshot ID (only if argument 3 has the value "3")
    5. Snapshot name (only if argument 3 has the value "3")

    For more details and script examples see:
    https://github.com/bit-team/user-callback

    Notes:
        The user-callback script file is normally implemented as
        shell script but could theoretically be implemented in
        any script language (declared via the hash bang "#!"
        in the first line of the script file.
    """
    def __init__(self):
        logger.openlog()
        return

    def init(self, snapshots):
        self.config = snapshots.config
        self.script = self.config.takeSnapshotUserCallback()

        return os.path.exists(self.script)

    # TODO 09/28/2022: This method should be private (_callback)
    def callback(self, *args, profileID=None):
        if profileID is None:
            profileID = self.config.currentProfile()

        profileName = self.config.profileName(profileID)
        cmd = [self.script, profileID, profileName]
        cmd.extend(str(x) for x in args)

        logger.debug(f'Call user-callback: {" ".join(cmd)}', self)

        stdout, stderr = PIPE, PIPE

        try:
            callback = Popen(cmd,
                             stdout=stdout,
                             stderr=stderr,
                             universal_newlines=True)
            output = callback.communicate()

            # Stdout
            if output[0]:
                logger.info("user-callback returned '"
                            + output[0].strip('\n') + "'",
                            self)

            # Stderr
            if output[1]:
                logger.error("user-callback returned '"
                             + output[1].strip('\n') + "'",
                             self)

            if callback.returncode != 0:
                logger.warning(
                    f'user-callback returncode: {callback.returncode}', self)
                raise StopException()

        except OSError as e:
            logger.error(
                f'Exception when trying to run user callback: {e.strerror}',
                self)

    def processBegin(self):
        self.callback('1')

    def processEnd(self):
        self.callback('2')

    def error(self, code, message):
        if not message:
            self.callback('4', code)
        else:
            self.callback('4', code, message)

    def newSnapshot(self, snapshot_id, snapshot_path):
        self.callback('3', snapshot_id, snapshot_path)

    def appStart(self):
        self.callback('5')

    def appExit(self):
        self.callback('6')

    def mount(self, profileID = None):
        self.callback('7', profileID = profileID)

    def unmount(self, profileID = None):
        self.callback('8', profileID = profileID)



# SPDX-FileCopyrightText: 2016 Taylor Raack
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import re
import subprocess
import sys
from test import generic
import json
import config
import version

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))


class BackInTime(generic.TestCase):
    def test_quiet_mode(self):
        output = subprocess.getoutput('python3 backintime.py --quiet')

        # Remove "WARNING:" and "ERROR:" messages from output.
        # They should appear even in quite-mode.
        output = filter(lambda line: not line.startswith('WARNING:')
                        and not line.startswith('ERROR:'),
                        output.split('\n'))
        output = '\n'.join(list(output))

        self.assertEqual('', output)

    def test_local_snapshot_is_successful(self):
        """From BIT initialization through snapshot

        From BIT initialization all the way through successful snapshot on a
        local mount. test one of the highest level interfaces a user could
        work with - the command line ensures that argument parsing,
        functionality, and output all work as expected is NOT intended to
        replace individual method tests, which are incredibly useful as well.

        Development notes (by Buhtz, 2023):
        Multiple tests do compare return codes and output on stdout. It is NOT
        tested what is on the file system. The intention might be a system
        test. But the asserts not qualified to answer the important questions
        and observe the intended behavior.  Heavy refactoring is needed. But
        because of the "level" of that tests it won't happen in the near
        future. Also maintenance costs of this tests are damn high because
        every tiny modification of BIT gives a false fail of this test.

        Development notes (by Buhtz, 2024-05):
        It is just dumb stdout parsing. I tend to remove this test because of
        the calculation of its value and its maintenance costs.
        """

        # ensure that we see full diffs of assert output if there are any
        self.maxDiff = None

        # create pristine source directory with single file
        subprocess.getoutput("chmod -R a+rwx /tmp/test && rm -rf /tmp/test")
        os.mkdir('/tmp/test')

        with open('/tmp/test/testfile', 'w') as f:
            f.write('some data')

        # create pristine snapshot directory
        subprocess.getoutput(
            "chmod -R a+rwx /tmp/snapshots && rm -rf /tmp/snapshots")
        os.mkdir('/tmp/snapshots')

        # remove restored directory
        subprocess.getoutput("rm -rf /tmp/restored")

        # install proper destination filesystem structure and verify output
        proc = subprocess.Popen(["./backintime",
                                 "--config",
                                 "test/config",
                                 "--share-path",
                                 self.sharePath,
                                 "check-config",
                                 # do not overwrite users crontab
                                 "--no-crontab"],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        output, error = proc.communicate()
        msg = 'Returncode: {}\nstderr: {}\nstdout: {}' \
              .format(proc.returncode, error.decode(), output.decode())

        self.assertEqual(proc.returncode, 0, msg)

        self.assertRegex(output.decode(), re.compile(r'''
Back In Time
Version: \d+.\d+.\d+.*

Back In Time comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions; type `backintime --license' for details.

(INFO: Update to config version \d+
)?
 \+--------------------------------\+
 |  Check/prepare snapshot path   |
 \+--------------------------------\+
Check/prepare snapshot path: done

 \+--------------------------------\+
 |          Check config          |
 \+--------------------------------\+
Check config: done

Config .*test/config profile 'Main profile' is fine.''', re.MULTILINE))

        # execute backup and verify output
        proc = subprocess.Popen(["./backintime",
                                 "--config", "test/config",
                                 "--share-path", self.sharePath,
                                 "backup"],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        output, error = proc.communicate()
        msg = 'Returncode: {}\nstderr: {}\nstdout: {}' \
              .format(proc.returncode, error.decode(), output.decode())
        self.assertEqual(proc.returncode, 0, msg)

        self.assertRegex(output.decode(), re.compile(r'''
Back In Time
Version: \d+.\d+.\d+.*

Back In Time comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions; type `backintime --license' for details.
''', re.MULTILINE))

        # Workaround until refactoring was done (Buhtz, Feb.'23)
        # The log output completely goes to stderr.
        # Note: DBus warnings at the begin and end are already ignored by the
        #       regex but if the BiT serviceHelper.py DBus daemon is not
        #       installed at all the warnings also occur in the middle of below
        #       expected INFO log lines so they are removed by filtering here.
        #       The same goes with Gtk warnings.

        line_beginnings_to_exclude = [
            "WARNING",
            "Warning",
        ]

        # Warnings currently known:
        # - "WARNING: D-Bus message:"
        # - "WARNING: Udev-based profiles cannot be changed or checked"
        # - "WARNING: Inhibit Suspend failed"
        # - "Warning: Ignoring XDG_SESSION_TYPE=wayland on Gnome. Use
        #    QT_QPA_PLATFORM=wayland to run on Wayland anyway"

        line_contains_to_exclude = [
            "Gtk-WARNING",
            "qt.qpa.plugin: Could not find the Qt platform plugin",
            'qt.dbus.integration: Could not connect "org.freedesktop.IBus" '
            'to globalEngineChanged(QString)',
        ]

        # remove lines via startswith()
        filtered_log_output = filter(
            lambda line: not any([
                line.startswith(ex) for ex in line_beginnings_to_exclude]),
            error.decode().split('\n')
        )

        # remove lines via __contains__()
        filtered_log_output = filter(
            lambda line: not any([
                ex in line for ex in line_contains_to_exclude]),
            filtered_log_output
        )

        # remove empty lines
        filtered_log_output = filter(
            lambda line: line,
            filtered_log_output
        )

        filtered_log_output = '\n'.join(filtered_log_output)

        self.assertRegex(filtered_log_output, re.compile(r'''INFO: Lock
INFO: Take a new snapshot. Profile: 1 Main profile
INFO: Call rsync to take the snapshot
INFO: Save config file
INFO: Save permissions
INFO: Unlock''', re.MULTILINE))

        # get snapshot id
        subprocess.check_output(["./backintime",
                                 "--config",
                                 "test/config",
                                 "snapshots-list"])

        # execute restore and verify output
        proc = subprocess.Popen(["./backintime",
                                 "--config",
                                 "test/config",
                                 "--share-path",
                                 self.sharePath,
                                 "restore",
                                 "/tmp/test/testfile",
                                 "/tmp/restored",
                                 "0"],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        output, error = proc.communicate()
        msg = 'Returncode: {}\nstderr: {}\nstdout: {}' \
              .format(proc.returncode, error.decode(), output.decode())
        self.assertEqual(proc.returncode, 0, msg)

        self.assertRegex(output.decode(), re.compile(r'''
Back In Time
Version: \d+.\d+.\d+.*

Back In Time comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions; type `backintime --license' for details.

''', re.MULTILINE))

        # The log output completely goes to stderr
        self.assertRegex(
            error.decode(),
            re.compile(
                r'''INFO: Restore: /tmp/test/testfile to: /tmp/restored.*''',
                re.MULTILINE
            )
        )

        # verify that files restored are the same as those backed up
        subprocess.check_output(["diff",
                                 "-r",
                                 "/tmp/test",
                                 "/tmp/restored"])

    def test_diagnostics_arg(self):
        # "output" from stdout may currently be polluted with logging output
        # lines from INFO and DEBUG log output.
        # Logging output of WARNING and ERROR is already written to stderr
        # so `check_output` does work here (returns only stdout without
        # stderr).
        output = subprocess.check_output(["./backintime", "--diagnostics"])

        diagnostics = json.loads(output)
        self.assertEqual(diagnostics["backintime"]["name"],
                         config.Config.APP_NAME)
        self.assertEqual(diagnostics["backintime"]["version"],
                         version.__version__)



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2008-2022 Taylor Raak
# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
from PyQt6.QtWidgets import (QDialog,
                             QVBoxLayout,
                             QHBoxLayout,
                             QLabel,
                             QCheckBox)
import config
import tools
import qttools
from manageprofiles import combobox


class OptionsTab(QDialog):
    """The 'Options' tab in the Manage Profiles dialog."""

    def __init__(self, parent):
        super().__init__(parent=parent)

        self._parent_dialog = parent

        tab_layout = QVBoxLayout(self)

        # layoutWidget = QWidget(self)
        # layout = QVBoxLayout(layoutWidget)

        self.cbNotify = QCheckBox(_('Enable notifications'), self)
        tab_layout.addWidget(self.cbNotify)

        self.cbNoSnapshotOnBattery \
            = QCheckBox(_('Disable snapshots when on battery'), self)
        tab_layout.addWidget(self.cbNoSnapshotOnBattery)

        if not tools.powerStatusAvailable():
            self.cbNoSnapshotOnBattery.setEnabled(False)
            self.cbNoSnapshotOnBattery.setToolTip(
                _('Power status not available from system'))

        self.cbGlobalFlock = QCheckBox(_('Run only one snapshot at a time'))
        tab_layout.addWidget(self.cbGlobalFlock)
        qttools.set_wrapped_tooltip(
            self.cbGlobalFlock,
            _('Other snapshots will be blocked until the current snapshot '
              'is done. This is a global option. So it will affect all '
              'profiles for this user. But you need to activate this for all '
              'other users, too.')
        )

        self.cbBackupOnRestore = QCheckBox(
            _('Backup replaced files on restore'), self)
        tab_layout.addWidget(self.cbBackupOnRestore)
        qttools.set_wrapped_tooltip(
            self.cbBackupOnRestore,
            _("Newer versions of files will be renamed with trailing {suffix} "
              "before restoring. If you don't need them anymore you can "
              "remove them with {cmd}").format(
                  suffix=self._parent_dialog.snapshots.backupSuffix(),
                  cmd='find ./ -name "*{suffix}" -delete'.format(
                      suffix=self._parent_dialog.snapshots.backupSuffix()
                  )
              )
        )

        self.cbContinueOnErrors = QCheckBox(
            _('Continue on errors (keep incomplete snapshots)'), self)
        tab_layout.addWidget(self.cbContinueOnErrors)

        self.cbUseChecksum = QCheckBox(
            _('Use checksum to detect changes'), self)
        tab_layout.addWidget(self.cbUseChecksum)

        self.cbTakeSnapshotRegardlessOfChanges = QCheckBox(
            _('Take a new snapshot whether there were changes or not.'))
        tab_layout.addWidget(self.cbTakeSnapshotRegardlessOfChanges)

        # log level
        hlayout = QHBoxLayout()
        tab_layout.addLayout(hlayout)

        hlayout.addWidget(QLabel(_('Log Level:'), self))

        self.comboLogLevel = self._combo_log_level()
        hlayout.addWidget(self.comboLogLevel)
        hlayout.addStretch()

        #
        tab_layout.addStretch()

    @property
    def config(self) -> config.Config:
        return self._parent_dialog.config

    def load_values(self):
        self.cbNotify.setChecked(self.config.notify())
        self.cbNoSnapshotOnBattery.setChecked(
            self.config.noSnapshotOnBattery())
        self.cbGlobalFlock.setChecked(self.config.globalFlock())
        self.cbBackupOnRestore.setChecked(self.config.backupOnRestore())
        self.cbContinueOnErrors.setChecked(self.config.continueOnErrors())
        self.cbUseChecksum.setChecked(self.config.useChecksum())
        self.cbTakeSnapshotRegardlessOfChanges.setChecked(
            self.config.takeSnapshotRegardlessOfChanges())
        self.comboLogLevel.select_by_data(self.config.logLevel())

    def store_values(self):
        self.config.setNotify(self.cbNotify.isChecked())
        self.config.setNoSnapshotOnBattery(
            self.cbNoSnapshotOnBattery.isChecked())
        self.config.setGlobalFlock(self.cbGlobalFlock.isChecked())
        self.config.setBackupOnRestore(self.cbBackupOnRestore.isChecked())
        self.config.setContinueOnErrors(self.cbContinueOnErrors.isChecked())
        self.config.setUseChecksum(self.cbUseChecksum.isChecked())
        self.config.setTakeSnapshotRegardlessOfChanges(
            self.cbTakeSnapshotRegardlessOfChanges.isChecked())
        self.config.setLogLevel(
            self.comboLogLevel.itemData(self.comboLogLevel.currentIndex()))

    def _combo_log_level(self):
        fill = {
            0: _('None'),
            1: _('Errors'),
            2: _('Changes') + ' & ' + _('Errors'),
            3: _('All'),
        }
        return combobox.BitComboBox(self, fill)



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import sys
import argparse
import atexit
import subprocess
from datetime import datetime
from time import sleep
import json
import pathlib
import tools
# Workaround for situations where startApp() is not invoked.
# E.g. when using --diagnostics and other argparse.Action
tools.initiate_translation(None)
import config
import logger
import snapshots
import sshtools
import mount
import password
import encfstools
import cli
from bitbase import URL_ENCRYPT_TRANSITION
from diagnostics import collect_diagnostics, collect_minimal_diagnostics
from exceptions import MountException
from applicationinstance import ApplicationInstance
from version import __version__

RETURN_OK = 0
RETURN_ERR = 1
RETURN_NO_CFG = 2

parsers = {}


def takeSnapshotAsync(cfg, checksum=False):
    """
    Fork a new backintime process with 'backup' command which will
    take a new snapshot in background.

    Args:
        cfg (config.Config): config that should be used
    """
    cmd = []

    if cfg.ioniceOnUser():
        cmd.extend(('ionice', '-c2', '-n7'))

    cmd.append('backintime')

    if '1' != cfg.currentProfile():
        cmd.extend(('--profile-id', str(cfg.currentProfile())))
    if cfg._LOCAL_CONFIG_PATH is not cfg._DEFAULT_CONFIG_PATH:
        cmd.extend(('--config', cfg._LOCAL_CONFIG_PATH))
    if cfg._LOCAL_DATA_FOLDER is not cfg._DEFAULT_LOCAL_DATA_FOLDER:
        cmd.extend(('--share-path', cfg.DATA_FOLDER_ROOT))
    if logger.DEBUG:
        cmd.append('--debug')
    if checksum:
        cmd.append('--checksum')

    cmd.append('backup')

    # child process need to start its own ssh-agent because otherwise
    # it would be lost without ssh-agent if parent will close
    env = os.environ.copy()

    for i in ('SSH_AUTH_SOCK', 'SSH_AGENT_PID'):
        try:
            del env[i]

        except:
            pass

    subprocess.Popen(cmd, env = env)


def takeSnapshot(cfg, force=True):
    """
    Take a new snapshot.

    Args:
        cfg (config.Config):    config that should be used
        force (bool):           take the snapshot even if it wouldn't need to
                                or would be prevented (e.g. running on battery)

    Returns:
        bool:                   ``True`` if there was an error
    """
    tools.envLoad(cfg.cronEnvFile())
    ret = snapshots.Snapshots(cfg).backup(force)

    return ret


def _mount(cfg):
    """
    Mount external filesystems.

    Args:
        cfg (config.Config): Config that should be used.
    """
    try:
        hash_id = mount.Mount(cfg=cfg).mount()

    except MountException as ex:
        logger.error(str(ex))
        sys.exit(RETURN_ERR)

    else:
        cfg.setCurrentHashId(hash_id)


def _umount(cfg):
    """
    Unmount external filesystems.

    Args:
        cfg (config.Config): Config that should be used.
    """
    try:
        mount.Mount(cfg=cfg).umount(cfg.current_hash_id)

    except MountException as ex:
        logger.error(str(ex))


def createParsers(app_name='backintime'):
    """
    Define parsers for commandline arguments.

    Args:
        app_name (str):         string representing the current application
    """
    global parsers

    #define debug
    debugArgsParser = argparse.ArgumentParser(add_help = False)
    debugArgsParser.add_argument('--debug',
                                 action = 'store_true',
                                 help = 'Increase verbosity.')

    #define config argument
    configArgsParser = argparse.ArgumentParser(add_help = False)
    configArgsParser.add_argument('--config',
                                 metavar = 'PATH',
                                 type = str,
                                 action = 'store',
                                 help = 'Read config from %(metavar)s. ' +
                                        'Default = ~/.config/backintime/config')

    configArgsParser.add_argument('--share-path',
                                 metavar = 'PATH',
                                 type = str,
                                 action = 'store',
                                 help = 'Write runtime data (locks, messages, log and mountpoints) to %(metavar)s.')

    #define common arguments which are used for all commands
    commonArgsParser = argparse.ArgumentParser(add_help = False, parents = [configArgsParser, debugArgsParser])
    profileGroup = commonArgsParser.add_mutually_exclusive_group()
    profileGroup.add_argument    ('--profile',
                                  metavar = 'NAME',
                                  type = str,
                                  action = 'store',
                                  help = 'Select profile by %(metavar)s.')
    profileGroup.add_argument    ('--profile-id',
                                  metavar = 'ID',
                                  type = int,
                                  action = 'store',
                                  help = 'Select profile by %(metavar)s.')
    commonArgsParser.add_argument('--quiet',
                                  action = 'store_true',
                                  help = 'Be quiet. Suppress messages on stdout.')

    #define arguments which are only used by snapshots-path, snapshots-list-path and last-snapshot-path
    snapshotPathParser = argparse.ArgumentParser(add_help = False)
    snapshotPathParser.add_argument('--keep-mount',
                                    action = 'store_true',
                                    help = "Don't unmount on exit.")

    #define arguments which are used by rsync commands (backup and restore)
    rsyncArgsParser = argparse.ArgumentParser(add_help = False)
    rsyncArgsParser.add_argument('--checksum',
                                 action = 'store_true',
                                 help = 'force to use checksum for checking if files have been changed.')

    #define arguments for snapshot remove
    removeArgsParser = argparse.ArgumentParser(add_help = False)
    removeArgsParser.add_argument('SNAPSHOT_ID',
                                  type = str,
                                  action = 'store',
                                  nargs = '*',
                                  help = 'ID of snapshots which should be removed.')

    #define main argument parser
    parser = argparse.ArgumentParser(prog = app_name,
                                     parents = [commonArgsParser],
                                     description = '%(app)s - a simple backup tool for GNU/Linux.'
                                                   % {'app': config.Config.APP_NAME},
                                     epilog = "For backwards compatibility commands can also be used with trailing '--'. "
                                              "All listed arguments will work with all commands. Some commands have extra arguments. "
                                              "Run '%(app_name)s <COMMAND> -h' to see the extra arguments."
                                              % {'app_name': app_name})
    parsers['main'] = parser
    parser.add_argument('--version', '-v',
                        action = 'version',
                        version = '%(prog)s ' + __version__,
                        help = "show %(prog)s's version number.")
    parser.add_argument('--license',
                        action = printLicense,
                        nargs = 0,
                        help = "show %(prog)s's license.")
    parser.add_argument('--diagnostics',
                        action = printDiagnostics,
                        nargs = 0,
                        help = "show helpful info for better support in case of issues (in JSON format)")

    #######################
    ### define commands ###
    #######################
    epilog = "Run '%(app_name)s -h' to get help for additional arguments. " %{'app_name': app_name}
    epilogCommon = epilog + 'Additional arguments: --config, --debug, --profile, --profile-id, --quiet'
    epilogConfig = epilog + 'Additional arguments: --config, --debug'

    subparsers = parser.add_subparsers(title = 'Commands', dest = 'command')
    command = 'backup'
    nargs = 0
    aliases = [(command, nargs), ('b', nargs)]
    description = 'Take a new snapshot. Ignore if the profile ' +\
                  'is not scheduled or if the machine is running on battery.'
    backupCP =             subparsers.add_parser(command,
                                                 parents = [rsyncArgsParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    backupCP.set_defaults(func = backup)
    parsers[command] = backupCP

    command = 'backup-job'
    nargs = 0
    aliases.append((command, nargs))
    description = 'Take a new snapshot in background only '      +\
                  'if the profile is scheduled and the machine ' +\
                  'is not on battery. This is used by cron jobs.'
    backupJobCP =          subparsers.add_parser(command,
                                                 parents = [rsyncArgsParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    backupJobCP.set_defaults(func = backupJob)
    parsers[command] = backupJobCP

    command = 'benchmark-cipher'
    nargs = '?'
    aliases.append((command, nargs))
    description = 'Show a benchmark of all ciphers for ssh transfer.'
    benchmarkCipherCP =    subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    benchmarkCipherCP.set_defaults(func = benchmarkCipher)
    parsers[command] = benchmarkCipherCP
    benchmarkCipherCP.add_argument              ('FILE_SIZE',
                                                 type = int,
                                                 action = 'store',
                                                 default = 40,
                                                 nargs = '?',
                                                 help = 'File size used for benchmark.')

    command = 'check-config'
    description = 'Check the profiles configuration and install crontab entries.'
    checkConfigCP =        subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    checkConfigCP.add_argument                  ('--no-crontab',
                                                 action = 'store_true',
                                                 help = 'Do not install crontab entries.')
    checkConfigCP.set_defaults(func = checkConfig)
    parsers[command] = checkConfigCP

    command = 'decode'
    nargs = '*'
    aliases.append((command, nargs))
    description = "Decode paths with 'encfsctl decode'"
    decodeCP =             subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    decodeCP.set_defaults(func = decode)
    parsers[command] = decodeCP
    decodeCP.add_argument                       ('PATH',
                                                 type = str,
                                                 action = 'store',
                                                 nargs = '*',
                                                 help = 'Decode PATH. If no PATH is specified on command line ' +\
                                                 'a list of filenames will be read from stdin.')

    command = 'last-snapshot'
    nargs = 0
    aliases.append((command, nargs))
    description = 'Show the ID of the last snapshot.'
    lastSnapshotCP =       subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    lastSnapshotCP.set_defaults(func = lastSnapshot)
    parsers[command] = lastSnapshotCP

    command = 'last-snapshot-path'
    nargs = 0
    aliases.append((command, nargs))
    description = 'Show the path of the last snapshot.'
    lastSnapshotsPathCP =  subparsers.add_parser(command,
                                                 parents = [snapshotPathParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    lastSnapshotsPathCP.set_defaults(func = lastSnapshotPath)
    parsers[command] = lastSnapshotsPathCP

    command = 'pw-cache'
    nargs = '*'
    aliases.append((command, nargs))
    description = 'Control Password Cache for non-interactive cronjobs.'
    pwCacheCP =            subparsers.add_parser(command,
                                                 epilog = epilogConfig,
                                                 help = description,
                                                 description = description)
    pwCacheCP.set_defaults(func = pwCache)
    parsers[command] = pwCacheCP
    pwCacheCP.add_argument                      ('ACTION',
                                                 action = 'store',
                                                 choices = ['start', 'stop', 'restart', 'reload', 'status'],
                                                 nargs = '?',
                                                 help = 'Command to send to Password Cache daemon.')

    command = 'remove'
    nargs = '*'
    aliases.append((command, nargs))
    description = 'Remove a snapshot.'
    removeCP =             subparsers.add_parser(command,
                                                 parents = [removeArgsParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    removeCP.set_defaults(func = remove)
    parsers[command] = removeCP

    command = 'remove-and-do-not-ask-again'
    nargs = '*'
    aliases.append((command, nargs))
    description = "Remove snapshots and don't ask for confirmation before. Be careful!"
    removeDoNotAskCP =     subparsers.add_parser(command,
                                                 parents = [removeArgsParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    removeDoNotAskCP.set_defaults(func = removeAndDoNotAskAgain)
    parsers[command] = removeDoNotAskCP

    command = 'restore'
    nargs = '*'
    aliases.append((command, nargs))
    description = 'Restore files.'
    restoreCP =            subparsers.add_parser(command,
                                                 parents = [rsyncArgsParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    restoreCP.set_defaults(func = restore)
    parsers[command] = restoreCP
    backupGroup = restoreCP.add_mutually_exclusive_group()
    restoreCP.add_argument                      ('WHAT',
                                                 type = str,
                                                 action = 'store',
                                                 nargs = '?',
                                                 help = 'Restore file or directory WHAT.')

    restoreCP.add_argument                      ('WHERE',
                                                 type = str,
                                                 action = 'store',
                                                 nargs = '?',
                                                 help = "Restore to WHERE. An empty argument '' will restore to original destination.")

    restoreCP.add_argument                      ('SNAPSHOT_ID',
                                                 type = str,
                                                 action = 'store',
                                                 nargs = '?',
                                                 help = 'Which SNAPSHOT_ID should be used. This can be a snapshot ID or ' +\
                                                 'an integer starting with 0 for the last snapshot, 1 for the second to last, ... ' +\
                                                 'the very first snapshot is -1')

    restoreCP.add_argument                      ('--delete',
                                                 action = 'store_true',
                                                 help = 'Restore and delete newer files which are not in the snapshot. ' +\
                                                 'WARNING: deleting files in filesystem root could break your whole system!!!')

    backupGroup.add_argument                    ('--local-backup',
                                                 action = 'store_true',
                                                 help = 'Create backup files before changing local files.')

    backupGroup.add_argument                    ('--no-local-backup',
                                                 action = 'store_true',
                                                 help = 'Temporarily disable creation of backup files before changing local files. ' +\
                                                 'This can be switched off permanently in Settings, too.')

    restoreCP.add_argument                      ('--only-new',
                                                 action = 'store_true',
                                                 help = 'Only restore files which do not exist or are newer than ' +\
                                                        'those in destination. Using "rsync --update" option.')

    command = 'shutdown'
    nargs = 0
    description = 'Shut down the computer after the snapshot is done.'
    shutdownCP =           subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    shutdownCP.set_defaults(func = shutdown)
    parsers[command] = shutdownCP

    command = 'smart-remove'
    nargs = 0
    description = 'Remove snapshots based on "Smart Removal" pattern.'
    smartRemoveCP =        subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    smartRemoveCP.set_defaults(func = smartRemove)
    parsers[command] = smartRemoveCP

    command = 'snapshots-list'
    nargs = 0
    aliases.append((command, nargs))
    description = 'Show a list of snapshot IDs.'
    snapshotsListCP =      subparsers.add_parser(command,
                                                 parents = [snapshotPathParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    snapshotsListCP.set_defaults(func = snapshotsList)
    parsers[command] = snapshotsListCP

    command = 'snapshots-list-path'
    nargs = 0
    aliases.append((command, nargs))
    description = "Show the paths to snapshots."
    snapshotsListPathCP =  subparsers.add_parser(command,
                                                 parents = [snapshotPathParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    snapshotsListPathCP.set_defaults(func = snapshotsListPath)
    parsers[command] = snapshotsListPathCP

    command = 'snapshots-path'
    nargs = 0
    aliases.append((command, nargs))
    description = 'Show the path where snapshots are stored.'
    snapshotsPathCP =      subparsers.add_parser(command,
                                                 parents = [snapshotPathParser],
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    snapshotsPathCP.set_defaults(func = snapshotsPath)
    parsers[command] = snapshotsPathCP

    command = 'unmount'
    nargs = 0
    aliases.append((command, nargs))
    description = 'Unmount the profile.'
    unmountCP =            subparsers.add_parser(command,
                                                 epilog = epilogCommon,
                                                 help = description,
                                                 description = description)
    unmountCP.set_defaults(func = unmount)
    parsers[command] = unmountCP

    #define aliases for all commands with trailing --
    group = parser.add_mutually_exclusive_group()
    for alias, nargs in aliases:
        if len(alias) == 1:
            arg = '-%s' % alias
        else:
            arg = '--%s' % alias

        group.add_argument(arg,
                           nargs=nargs,
                           action=PseudoAliasAction,
                           help=argparse.SUPPRESS)


def encfs_deprecation_warning():
    """Warn about encfs deprecation in syslog.

    See Issue #1734 for details. This function is a workraound and will be
    removed if #1734 is closed.
    """

    # Don't warn if EncFS isn't installed
    if not tools.checkCommand('encfs'):
        return

    # Timestamp file
    xdg_state = os.environ.get('XDG_STATE_HOME', None)
    if xdg_state:
        xdg_state = pathlib.Path(xdg_state)
    else:
        xdg_state = pathlib.Path.home() / '.local' / 'state'
    fp = xdg_state / 'backintime.encfs-warning.timestamp'

    # ensure existence
    if not fp.exists():
        fp.parent.mkdir(parents=True, exist_ok=True)
        fp.touch()

    # Calculate age of that file
    delta = datetime.now() - datetime.fromtimestamp(fp.stat().st_mtime)

    # Don't warn if to young
    if delta.days < 30:
        return

    logger.warning('EncFS encrypted profiles are deprecated in Back In Time. '
                   'Removal is schedule for minor release 1.7 in year 2026. '
                   'For details and alternatives '
                   f'read: {URL_ENCRYPT_TRANSITION}')


    # refresh timestamp
    fp.touch()


def startApp(app_name='backintime'):
    """
    Start the requested command or return config if there was no command
    in arguments.

    Args:
        app_name (str): string representing the current application

    Returns:
        config.Config:  current config if no command was given in arguments
    """
    createParsers(app_name)

    logger.openlog()

    args = argParse(None)

    # Name, Version, As Root, OS
    msg = ''
    for key, val in collect_minimal_diagnostics().items():
        msg = f'{msg}; {key}: {val}'
    logger.debug(msg[2:])

    # Add source path to $PATH environ if running from source
    if tools.runningFromSource():
        tools.addSourceToPathEnviron()

    # Warn about sudo
    if (tools.usingSudo()
            and os.getenv('BIT_SUDO_WARNING_PRINTED', 'false') == 'false'):

        os.putenv('BIT_SUDO_WARNING_PRINTED', 'true')
        logger.warning(
            "It looks like you're using 'sudo' to start "
            f"{config.Config.APP_NAME}. This will cause some trouble. "
            f"Please use either 'sudo -i {app_name}' or 'pkexec {app_name}'.")

    encfs_deprecation_warning()

    # Call commands
    if 'func' in dir(args):
        args.func(args)

    else:
        setQuiet(args)
        printHeader()

        return getConfig(args, False)


def argParse(args):
    """
    Parse arguments given on commandline.

    Args:
        args (argparse.Namespace):  Namespace that should be enhanced
                                    or ``None``

    Returns:
        argparser.Namespace:        new parsed Namespace
    """
    def join(args, subArgs):
        """
        Add new arguments to existing Namespace.

        Args:
            args (argparse.Namespace):
                        main Namespace that should get new arguments
            subArgs (argparse.Namespace):
                        second Namespace which have new arguments
                        that should be merged into ``args``
        """
        for key, value in vars(subArgs).items():
            # Only add new values if it isn't set already or if there really IS
            # a value
            if getattr(args, key, None) is None or value:
                setattr(args, key, value)

    # First parse the main parser without subparsers
    # otherwise positional args in subparsers will be to greedy
    # but only if -h or --help is not involved because otherwise
    # help will not work for subcommands
    mainParser = parsers['main']
    sub = []

    if '-h' not in sys.argv and '--help' not in sys.argv:

        for i in mainParser._actions:

            if isinstance(i, argparse._SubParsersAction):
                # Remove subparsers
                mainParser._remove_action(i)
                sub.append(i)

    args, unknownArgs = mainParser.parse_known_args(args)

    # Read subparsers again
    if sub:
        [mainParser._add_action(i) for i in sub]

    # Parse it again for unknown args
    if unknownArgs:
        subArgs, unknownArgs = mainParser.parse_known_args(unknownArgs)
        join(args, subArgs)

    # Finally parse only the command parser, otherwise we miss some arguments
    # from command
    if unknownArgs and 'command' in args and args.command in parsers:
        commandParser = parsers[args.command]
        subArgs, unknownArgs = commandParser.parse_known_args(unknownArgs)
        join(args, subArgs)

    try:
        logger.DEBUG = args.debug
    except AttributeError:
        pass

    args_dict = vars(args)
    used_args = {
        key: args_dict[key]
        for key
        in filter(lambda key: args_dict[key] is not None, args_dict)
    }

    logger.debug(f'Argument(s) used: {used_args}')

    # Report unknown arguments but not if we run aliasParser next because we
    # will parse again in there.
    if unknownArgs and not ('func' in args and args.func is aliasParser):
        mainParser.error(f'Unknown argument(s): {unknownArgs}')

    return args


def printHeader():
    """
    Print application name, version and legal notes.
    """
    print('')
    print('Back In Time')
    print('Version: ' + __version__)
    print('')
    print('Back In Time comes with ABSOLUTELY NO WARRANTY.')
    print('This is free software, and you are welcome to redistribute it')
    print("under certain conditions; type `backintime --license' for details.")
    print('')


class PseudoAliasAction(argparse.Action):
    """
    Translate '--COMMAND' into 'COMMAND' for backwards compatibility.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        """
        Translate '--COMMAND' into 'COMMAND' for backwards compatibility.

        Args:
            parser (argparse.ArgumentParser): NotImplemented
            namespace (argparse.Namespace):   Namespace that should get modified
            values:                           NotImplemented
            option_string:                    NotImplemented
        """
        #TODO: find a more elegant way to solve this
        dest = self.dest.replace('_', '-')
        if self.dest == 'b':
            replace = '-b'
            alias = 'backup'
        else:
            replace = '--%s' % dest
            alias = dest
        setattr(namespace, 'func', aliasParser)
        setattr(namespace, 'replace', replace)
        setattr(namespace, 'alias', alias)


def aliasParser(args):
    """
    Call commands which where given with leading -- for backwards
    compatibility.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments
    """
    if not args.quiet:
        logger.info("Run command '%(alias)s' instead of argument '%(replace)s' due to backwards compatibility."
                    % {'alias': args.alias, 'replace': args.replace})
    argv = [w.replace(args.replace, args.alias) for w in sys.argv[1:]]
    newArgs = argParse(argv)
    if 'func' in dir(newArgs):
        newArgs.func(newArgs)


def getConfig(args, check=True):
    """
    Load config and change to profile selected on commandline.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments
        check (bool):   if ``True`` check if config is valid

    Returns:
        config.Config:  current config with requested profile selected

    Raises:
        SystemExit:     1 if ``profile`` or ``profile_id`` is no valid profile
                        2 if ``check`` is ``True`` and config is not configured
    """
    cfg = config.Config(config_path=args.config, data_path=args.share_path)
    logger.debug('config file: "{}"; share path: "{}"; profiles: "{}"'.format(
        cfg._LOCAL_CONFIG_PATH,
        cfg._LOCAL_DATA_FOLDER,
        ', '.join(f'{profile_id}={cfg.profileName(profile_id)}'
                  for profile_id in cfg.profiles())
    ))

    if 'profile_id' in args and args.profile_id:
        if not cfg.setCurrentProfile(args.profile_id):
            logger.error('Profile-ID not found: %s' % args.profile_id)
            sys.exit(RETURN_ERR)

    if 'profile' in args and args.profile:
        if not cfg.setCurrentProfileByName(args.profile):
            logger.error('Profile not found: %s' % args.profile)
            sys.exit(RETURN_ERR)

    if check and not cfg.isConfigured():
        logger.error('%(app)s is not configured!' % {'app': cfg.APP_NAME})
        sys.exit(RETURN_NO_CFG)

    if 'checksum' in args:
        cfg.forceUseChecksum = args.checksum

    return cfg


def setQuiet(args):
    """
    Redirect :py:data:`sys.stdout` to ``/dev/null`` if ``--quiet`` was set on
    commandline. Return the original :py:data:`sys.stdout` file object which can
    be used to print absolute necessary information.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Returns:
        sys.stdout:     default sys.stdout
    """
    force_stdout = sys.stdout
    if args.quiet:
        # do not replace with subprocess.DEVNULL - will not work
        sys.stdout = open(os.devnull, 'w')
        atexit.register(sys.stdout.close)
        atexit.register(force_stdout.close)
    return force_stdout


class printLicense(argparse.Action):
    """
    Print custom license
    """

    def __init__(self, *args, **kwargs):
        super(printLicense, self).__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        license_path = pathlib.Path(tools.docPath()) / 'LICENSE'
        print(license_path.read_text('utf-8'))
        sys.exit(RETURN_OK)


class printDiagnostics(argparse.Action):
    """
    Print information that is helpful for the support team
    to narrow down problems and bugs.
    The info is printed using the machine- and human-readable JSON format
    """

    def __init__(self, *args, **kwargs):
        super(printDiagnostics, self).__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):

        diagnostics = collect_diagnostics()

        print(json.dumps(diagnostics, indent=4))

        sys.exit(RETURN_OK)


def backup(args, force=True):
    """
    Command for force taking a new snapshot.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments
        force (bool):   take the snapshot even if it wouldn't need to or would
                        be prevented (e.g. running on battery)

    Raises:
        SystemExit:     0 if successful, 1 if not
    """
    setQuiet(args)
    printHeader()
    cfg = getConfig(args)
    ret = takeSnapshot(cfg, force)
    sys.exit(int(ret))


def backupJob(args):
    """
    Command for taking a new snapshot in background. Mainly used for cronjobs.
    This will run the snapshot inside a daemon and detach from it. It will
    return immediately back to commandline.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    cli.BackupJobDaemon(backup, args).start()


def shutdown(args):
    """
    Command for shutting down the computer after the current snapshot has
    finished.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0 if successful; 1 if it failed either because there is
                        no active snapshot for this profile or shutdown is not
                        supported.
    """
    setQuiet(args)
    printHeader()
    cfg = getConfig(args)

    sd = tools.ShutDown()

    if not sd.canShutdown():
        logger.warning('Shutdown is not supported.')
        sys.exit(RETURN_ERR)

    instance = ApplicationInstance(cfg.takeSnapshotInstanceFile(), False)
    profile = '='.join((cfg.currentProfile(), cfg.profileName()))

    if not instance.busy():
        logger.info('There is no active snapshot for profile %s. Skip shutdown.'
                    %profile)
        sys.exit(RETURN_ERR)

    print('Shutdown is waiting for the snapshot in profile %s to end.\nPress CTRL+C to interrupt shutdown.\n'
          %profile)
    sd.activate_shutdown = True

    try:
        while instance.busy():
            logger.debug('Snapshot is still active. Wait for shutdown.')
            sleep(5)

    except KeyboardInterrupt:
        print('Shutdown interrupted.')

    else:
        logger.info('Shutdown now.')
        sd.shutdown()

    sys.exit(RETURN_OK)


def snapshotsPath(args):
    """
    Command for printing the full snapshot path of current profile.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    force_stdout = setQuiet(args)
    cfg = getConfig(args)
    if args.keep_mount:
        _mount(cfg)
    if args.quiet:
        msg = '{}'
    else:
        msg = 'SnapshotsPath: {}'
    print(msg.format(cfg.snapshotsFullPath()), file=force_stdout)
    sys.exit(RETURN_OK)


def snapshotsList(args):
    """
    Command for printing a list of all snapshots in current profile.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    force_stdout = setQuiet(args)
    cfg = getConfig(args)
    _mount(cfg)

    if args.quiet:
        msg = '{}'
    else:
        msg = 'SnapshotID: {}'
    no_sids = True
    #use snapshots.listSnapshots instead of iterSnapshots because of sorting
    for sid in snapshots.listSnapshots(cfg, reverse = False):
        print(msg.format(sid), file=force_stdout)
        no_sids = False
    if no_sids:
        logger.error("There are no snapshots in '%s'" % cfg.profileName())
    if not args.keep_mount:
        _umount(cfg)
    sys.exit(RETURN_OK)


def snapshotsListPath(args):
    """
    Command for printing a list of all snapshots paths in current profile.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    force_stdout = setQuiet(args)
    cfg = getConfig(args)
    _mount(cfg)

    if args.quiet:
        msg = '{}'
    else:
        msg = 'SnapshotPath: {}'
    no_sids = True
    #use snapshots.listSnapshots instead of iterSnapshots because of sorting
    for sid in snapshots.listSnapshots(cfg, reverse = False):
        print(msg.format(sid.path()), file=force_stdout)
        no_sids = False
    if no_sids:
        logger.error("There are no snapshots in '%s'" % cfg.profileName())
    if not args.keep_mount:
        _umount(cfg)
    sys.exit(RETURN_OK)


def lastSnapshot(args):
    """
    Command for printing the very last snapshot in current profile.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    force_stdout = setQuiet(args)
    cfg = getConfig(args)
    _mount(cfg)
    sid = snapshots.lastSnapshot(cfg)
    if sid:
        if args.quiet:
            msg = '{}'
        else:
            msg = 'SnapshotID: {}'
        print(msg.format(sid), file=force_stdout)
    else:
        logger.error("There are no snapshots in '%s'" % cfg.profileName())
    _umount(cfg)
    sys.exit(RETURN_OK)


def lastSnapshotPath(args):
    """
    Command for printing the path of the very last snapshot in
    current profile.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    force_stdout = setQuiet(args)
    cfg = getConfig(args)
    _mount(cfg)
    sid = snapshots.lastSnapshot(cfg)
    if sid:
        if args.quiet:
            msg = '{}'
        else:
            msg = 'SnapshotPath: {}'
        print(msg.format(sid.path()), file=force_stdout)
    else:
        logger.error("There are no snapshots in '%s'" % cfg.profileName())
    if not args.keep_mount:
        _umount(cfg)
    sys.exit(RETURN_OK)


def unmount(args):
    """
    Command for unmounting all filesystems.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    setQuiet(args)
    cfg = getConfig(args)
    _mount(cfg)
    _umount(cfg)
    sys.exit(RETURN_OK)


def benchmarkCipher(args):
    """
    Command for transferring a file with scp to remote host with all
    available ciphers and print its speed and time.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    setQuiet(args)
    printHeader()

    cfg = getConfig(args)

    if cfg.snapshotsMode() in ('ssh', 'ssh_encfs'):
        ssh = sshtools.SSH(cfg)
        ssh.benchmarkCipher(args.FILE_SIZE)
        sys.exit(RETURN_OK)

    else:
        logger.error("SSH is not configured for profile '%s'!" % cfg.profileName())
        sys.exit(RETURN_ERR)


def pwCache(args):
    """
    Command for starting password cache daemon.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0 if daemon is running, 1 if not
    """
    force_stdout = setQuiet(args)
    printHeader()

    cfg = getConfig(args)
    ret = RETURN_OK
    daemon = password.Password_Cache(cfg)

    if args.ACTION and args.ACTION != 'status':
        getattr(daemon, args.ACTION)()

    elif args.ACTION == 'status':

        print('%(app)s Password Cache: ' % {'app': cfg.APP_NAME},
              end=' ',
              file=force_stdout)

        if daemon.status():
            print(cli.bcolors.OKGREEN + 'running' + cli.bcolors.ENDC,
                  file=force_stdout)
            ret = RETURN_OK

        else:
            print(cli.bcolors.FAIL + 'not running' + cli.bcolors.ENDC,
                  file=force_stdout)
            ret = RETURN_ERR

    else:
        daemon.run()

    sys.exit(ret)


def decode(args):
    """
    Command for decoding paths given paths with 'encfsctl'.
    Will listen on stdin if no path was given.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    force_stdout = setQuiet(args)
    cfg = getConfig(args)

    if cfg.snapshotsMode() not in ('local_encfs', 'ssh_encfs'):
        logger.error("Profile '%s' is not encrypted." % cfg.profileName())
        sys.exit(RETURN_ERR)

    _mount(cfg)
    d = encfstools.Decode(cfg)

    if not args.PATH:

        while True:

            try:
                path = input()
            except EOFError:
                break

            if not path:
                break

            print(d.path(path), file=force_stdout)

    else:
        print('\n'.join(d.list(args.PATH)), file=force_stdout)

    d.close()
    _umount(cfg)

    sys.exit(RETURN_OK)


def remove(args, force=False):
    """
    Command for removing snapshots.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments
        force (bool):   don't ask before removing (BE CAREFUL!)

    Raises:
        SystemExit:     0
    """
    setQuiet(args)
    printHeader()

    cfg = getConfig(args)
    _mount(cfg)

    cli.remove(cfg, args.SNAPSHOT_ID, force)
    _umount(cfg)

    sys.exit(RETURN_OK)


def removeAndDoNotAskAgain(args):
    """
    Command for removing snapshots without asking before remove
    (BE CAREFUL!)

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    remove(args, True)


def smartRemove(args):
    """
    Command for running Smart-Removal from Terminal.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0 if okay
                        2 if Smart-Removal is not configured
    """
    setQuiet(args)
    printHeader()
    cfg = getConfig(args)
    sn = snapshots.Snapshots(cfg)

    enabled, keep_all, keep_one_per_day, keep_one_per_week, keep_one_per_month = cfg.smartRemove()
    if enabled:
        _mount(cfg)
        del_snapshots = sn.smartRemoveList(datetime.today(),
                                           keep_all,
                                           keep_one_per_day,
                                           keep_one_per_week,
                                           keep_one_per_month)
        logger.info('Smart Removal will remove {} snapshots'.format(len(del_snapshots)))
        sn.smartRemove(del_snapshots, log = logger.info)
        _umount(cfg)
        sys.exit(RETURN_OK)
    else:
        logger.error('Smart Removal is not configured.')
        sys.exit(RETURN_NO_CFG)


def restore(args):
    """
    Command for restoring files from snapshots.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0
    """
    setQuiet(args)
    printHeader()
    cfg = getConfig(args)
    _mount(cfg)

    if cfg.backupOnRestore() and not args.no_local_backup:
        backup = True
    else:
        backup = args.local_backup

    cli.restore(cfg,
                args.SNAPSHOT_ID,
                args.WHAT,
                args.WHERE,
                delete=args.delete,
                backup=backup,
                only_new=args.only_new)

    _umount(cfg)

    sys.exit(RETURN_OK)


def checkConfig(args):
    """
    Command for checking the config file.

    Args:
        args (argparse.Namespace):
                        previously parsed arguments

    Raises:
        SystemExit:     0 if config is okay, 1 if not
    """
    force_stdout = setQuiet(args)
    printHeader()
    cfg = getConfig(args)

    if cli.checkConfig(cfg, crontab=not args.no_crontab):
        print("\nConfig %(cfg)s profile '%(profile)s' is fine."
              % {'cfg': cfg._LOCAL_CONFIG_PATH,
                 'profile': cfg.profileName()},
              file=force_stdout)
        sys.exit(RETURN_OK)

    else:
        print("\nConfig %(cfg)s profile '%(profile)s' has errors."
              % {'cfg': cfg._LOCAL_CONFIG_PATH,
                 'profile': cfg.profileName()},
              file=force_stdout)
        sys.exit(RETURN_ERR)


if __name__ == '__main__':
    startApp()



# SPDX-FileCopyrightText: © 2024 Christian Buhtz <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Management of the state file."""
# pylint: disable=wrong-import-position,wrong-import-order
from __future__ import annotations
import os
import json
from pathlib import Path
from datetime import datetime, timezone
from copy import deepcopy
from qttools_path import registerBackintimePath
registerBackintimePath('common')
import singleton  # noqa: E402
import logger  # noqa: E402
import tools  # noqa: E402
from version import __version__  # noqa: E402


class StateData(dict, metaclass=singleton.Singleton):
    """Manage state data for Back In Time.

    Dev note (buhtz, 2024-12): It is usually recommended and preferred to
    derive from `collections.UserDict` instead of just `dict`. But this
    conflicts with the ``metaclass=``. To my current knowledge this is not a
    big deal and won't introduce any problems.

    """
    # pylint: disable=too-many-instance-attributes
    # The default structure. All properties do rely on them and assuming
    # it is there.
    _EMPTY_STRUCT = {
        'gui': {
            'mainwindow': {
                'files_view': {},
                'last_path': {},
                'places_sorting': {},
            },
            'manage_profiles': {
                'incl_sorting': {},
                'excl_sorting': {},
            },
            'logview': {},
        },
        'message': {
            'encfs': {}
        },
    }

    class Profile:
        """A surrogate to access profile-specific state data."""

        def __init__(self, profile_id: str, state: StateData):
            self._state = state
            self._profile_id = profile_id

        @property
        def msg_encfs(self) -> int:
            """Stage of EncFS deprecation warning shown as last."""
            try:
                return self._state['message']['encfs'][self._profile_id]
            except KeyError:
                self.msg_encfs = 0
                return self.msg_encfs

        @msg_encfs.setter
        def msg_encfs(self, val: int) -> None:
            self._state['message']['encfs'][self._profile_id] = val

        @property
        def last_path(self) -> Path:
            """Last path used in the GUI.

            Raises:
                KeyError
            """
            return Path(self._state['gui']['mainwindow'][
                'last_path'][self._profile_id])

        @last_path.setter
        def last_path(self, path: Path) -> None:
            self._state['gui']['mainwindow'][
                'last_path'][self._profile_id] = str(path)

        @property
        def places_sorting(self) -> tuple[int, int]:
            """Column index and sort order.

            Returns:
                Tuple with column index and its sorting order (0=ascending).
            """
            return self._state['gui']['mainwindow'][
                'places_sorting'][self._profile_id]

        @places_sorting.setter
        def places_sorting(self, vals: tuple[int, int]) -> None:
            self._state['gui']['mainwindow'][
                'places_sorting'][self._profile_id] = vals

        @property
        def exclude_sorting(self) -> tuple[int, int]:
            """Column index and sort order.

            Returns:
                Tuple with column index and its sorting order (0=ascending).
            """
            return self._state['gui']['manage_profiles'][
                'excl_sorting'][self._profile_id]

        @exclude_sorting.setter
        def exclude_sorting(self, vals: tuple[int, int]) -> None:
            self._state['gui']['manage_profiles'][
                'excl_sorting'][self._profile_id] = vals

        @property
        def include_sorting(self) -> tuple[int, int]:
            """Column index and sort order.

            Returns:
                Tuple with column index and its sorting order (0=ascending).
            """
            return self._state['gui']['manage_profiles'][
                'incl_sorting'][self._profile_id]

        @include_sorting.setter
        def include_sorting(self, vals: tuple[int, int]) -> None:
            self._state['gui']['manage_profiles'][
                'incl_sorting'][self._profile_id] = vals

    @staticmethod
    def file_path() -> Path:
        """Returns the state file path."""
        xdg_state = os.environ.get('XDG_STATE_HOME', None)
        if xdg_state:
            xdg_state = Path(xdg_state)
        else:
            xdg_state = Path.home() / '.local' / 'state'

        fp = xdg_state / 'backintime-qt.json'

        logger.debug(f'State file path: {fp}')

        return fp

    def __init__(self, data: dict = None):
        """Constructor."""

        # default
        full = deepcopy(self._EMPTY_STRUCT)

        if data:
            full = tools.nested_dict_update(full, data)

        super().__init__(full)

    def __str__(self):
        return json.dumps(self, indent=4)

    def _set_save_meta_data(self):
        meta = {
            'saved': datetime.now().isoformat(),
            'saved_utc': datetime.now(timezone.utc).isoformat(),
            'bitversion': __version__,
        }

        self['_meta'] = meta

    def save(self):
        """Store application state data to a file."""
        logger.debug('Save state data.')

        self._set_save_meta_data()

        fp = self.file_path()
        fp.parent.mkdir(parents=True, exist_ok=True)

        with fp.open('w', encoding='utf-8') as handle:
            handle.write(str(self))

    def profile(self, profile_id: str) -> StateData.Profile:
        """Return a `Profile` object related to the given id.

        Args:
            profile_id: A profile_id of a snapshot profile.

        Returns:
            A profile surrogate.

        Raises:
            KeyError: If profile does not exists.
        """
        return StateData.Profile(profile_id=profile_id, state=self)

    def manual_starts_countdown(self) -> int:
        """Countdown value about how often the users started the Back In Time
        GUI.

        At the end of the countown the `ApproachTranslatorDialog` is presented
        to the user.
        """
        return self.get('manual_starts_countdown', 10)

    def decrement_manual_starts_countdown(self):
        """Counts down to -1.

        See :py:func:`manual_starts_countdown()` for details.
        """
        val = self.manual_starts_countdown()

        if val > -1:
            self['manual_starts_countdown'] = val - 1

    @property
    def msg_release_candidate(self) -> str:
        """Last version of Back In Time in which the release candidate message
        box was displayed.
        """
        try:
            return self['message']['release_candidate']
        except KeyError:
            self.msg_release_candidate = None
            return self.msg_release_candidate

    @msg_release_candidate.setter
    def msg_release_candidate(self, val: str) -> None:
        self['message']['release_candidate'] = val

    @property
    def msg_encfs_global(self) -> int:
        """Last stage of global EncFS deprecation message that was shown."""
        try:
            return self['message']['encfs']['global']
        except KeyError:
            self.msg_encfs_global = 0
            return self.msg_encfs_global

    @msg_encfs_global.setter
    def msg_encfs_global(self, val: int) -> None:
        self['message']['encfs']['global'] = val

    @property
    def mainwindow_show_hidden(self) -> bool:
        """Show hidden files in files view."""
        try:
            return self['gui']['mainwindow']['show_hidden']
        except KeyError:
            self.mainwindow_show_hidden = False
            return self.mainwindow_show_hidden

    @mainwindow_show_hidden.setter
    def mainwindow_show_hidden(self, val: bool) -> None:
        self['gui']['mainwindow']['show_hidden'] = val

    @property
    def mainwindow_dims(self) -> tuple[int, int]:
        """Dimensions of the main window.

        Raises:
            KeyError
        """
        return self['gui']['mainwindow']['dims']

    @mainwindow_dims.setter
    def mainwindow_dims(self, vals: tuple[int, int]) -> None:
        self['gui']['mainwindow']['dims'] = vals

    @property
    def mainwindow_coords(self) -> tuple[int, int]:
        """Coordinates (position) of the main window.

        Raises:
            KeyError
        """
        return self['gui']['mainwindow']['coords']

    @mainwindow_coords.setter
    def mainwindow_coords(self, vals: tuple[int, int]) -> None:
        self['gui']['mainwindow']['coords'] = vals

    @property
    def logview_dims(self) -> tuple[int, int]:
        """Dimensions of the log view dialog.

        Raises:
            KeyError
        """
        try:
            return self['gui']['logview']['dims']
        except KeyError:
            self.logview_dims = (800, 500)
            return self.logview_dims

    @logview_dims.setter
    def logview_dims(self, vals: tuple[int, int]) -> None:
        self['gui']['logview']['dims'] = vals

    @property
    def files_view_sorting(self) -> tuple[int, int]:
        """Column index and sort order.

        Returns:
            Tuple with column index and its sorting order (0=ascending).
        """
        try:
            return self['gui']['mainwindow']['files_view']['sorting']
        except KeyError:
            self.files_view_sorting = (0, 0)
            return self.files_view_sorting

    @files_view_sorting.setter
    def files_view_sorting(self, vals: tuple[int, int]) -> None:
        self['gui']['mainwindow']['files_view']['sorting'] = vals

    @property
    def files_view_col_widths(self) -> tuple:
        """Widths of columns in the files view."""
        return self['gui']['mainwindow']['files_view']['col_widths']

    @files_view_col_widths.setter
    def files_view_col_widths(self, widths: tuple) -> None:
        self['gui']['mainwindow']['files_view']['col_widths'] = widths

    @property
    def mainwindow_main_splitter_widths(self) -> tuple[int, int]:
        """Left and right width of main splitter in main window.

        Returns:
            Two entry tuple with right and left widths.
        """
        try:
            return self['gui']['mainwindow']['splitter_main_widths']
        except KeyError:
            self.mainwindow_main_splitter_widths = (150, 450)
            return self.mainwindow_main_splitter_widths

    @mainwindow_main_splitter_widths.setter
    def mainwindow_main_splitter_widths(self, vals: tuple[int, int]) -> None:
        self['gui']['mainwindow']['splitter_main_widths'] = vals

    @property
    def mainwindow_second_splitter_widths(self) -> tuple[int, int]:
        """Left and right width of second splitter in main window.

        Returns:
            Two entry tuple with right and left widths.
        """
        try:
            return self['gui']['mainwindow']['splitter_second_widths']
        except KeyError:
            self.mainwindow_second_splitter_widths = (150, 300)
            return self.mainwindow_second_splitter_widths

    @mainwindow_second_splitter_widths.setter
    def mainwindow_second_splitter_widths(self, vals: tuple[int, int]) -> None:
        self['gui']['mainwindow']['splitter_second_widths'] = vals

    @property
    def toolbar_button_style(self) -> int:
        """Style of icons for the main toolbar.

        Returns:
           Style value as integer (default: 0 as ``ToolButtonIconOnly``)
        """
        try:
            return self['gui']['mainwindow']['toolbar_button_style']
        except KeyError:
            self.toolbar_button_style = 0
            return self.toolbar_button_style

    @toolbar_button_style.setter
    def toolbar_button_style(self, value) -> None:
        self['gui']['mainwindow']['toolbar_button_style'] = value



# SPDX-FileCopyrightText: © 2023 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Dialog window to choose GUI display language."""
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (QApplication,
                             QDialog,
                             QWidget,
                             QScrollArea,
                             QGridLayout,
                             QVBoxLayout,
                             QDialogButtonBox,
                             QRadioButton,
                             )
import tools
import languages
import qttools

LOW_RESOLUTION_WIDTH = 1024


class LanguageDialog(QDialog):
    """Dialog to choose GUI language."""
    def __init__(self, used_language_code: str, configured_language_code: str):
        super().__init__()

        self.language_code = None
        self.used_language_code = used_language_code
        self.configured_language_code = configured_language_code

        self.setWindowTitle(_('Setup language'))
        self.setWindowFlag(Qt.WindowType.WindowMaximizeButtonHint, True)

        scroll = QScrollArea(self)
        scroll.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        scroll.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        scroll.setWidget(self._language_widget())
        self._scroll = scroll

        # Fit the width of scrollarea to its content
        new_width = self._calculate_scroll_area_width()
        self._scroll.setMinimumWidth(new_width)

        buttonbox = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Cancel |
            QDialogButtonBox.StandardButton.Ok,
            self)

        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)

        layout = QVBoxLayout(self)
        layout.addWidget(scroll)
        layout.addWidget(buttonbox)

    def _calculate_scroll_area_width(self):
        """Credits:
         - https://stackoverflow.com/a/9081579/4865723
         - https://stackoverflow.com/a/76738806/4865723
        """
        widget_width = self._scroll.widget().sizeHint().width()
        scrollbar_width = self._scroll.verticalScrollBar().sizeHint().width()

        return widget_width + scrollbar_width

    def _create_radio_button(self, lang_code, label, tooltip) -> QRadioButton:
        r = QRadioButton(label, self)
        r.setToolTip(tooltip)
        r.toggled.connect(self.slot_radio)
        r.lang_code = lang_code

        # Is it the current used AND configured language?
        if (r.lang_code == self.used_language_code
                and r.lang_code == self.configured_language_code):

            r.setChecked(True)

        # "System default"
        elif self.configured_language_code == '' and r.lang_code is None:
            r.setChecked(True)

        return r

    def _language_widget(self):
        grid = QGridLayout()
        widget = QWidget(self)
        widget.setLayout(grid)

        # Entry: System default language
        label = 'System default'
        translated_label = _('System default')

        # If translation for that term exists...
        if label != translated_label:
            # ...combine source and translated version.
            label = f'{label}\n{translated_label}'

        r = self._create_radio_button(
            lang_code=None,
            label=label,
            tooltip=_('Use operating systems language.')
        )
        grid.addWidget(r, 1, 1)

        # Sort by language code but keep English on top
        langs = tools.get_language_names(self.used_language_code)
        sorted_codes = sorted(langs.keys())
        sorted_codes.remove('en')
        sorted_codes = ['en'] + sorted_codes

        # Number of columns used for radio buttons
        number_of_columns = 3

        # Low-resolution screens (XGA or less)
        if QApplication.primaryScreen().size().width() <= LOW_RESOLUTION_WIDTH:
            # Use one columns less
            number_of_columns -= 1

        # Calculate number of entries (rows) per column
        per_col_n = len(sorted_codes) / number_of_columns
        per_col_n = int(per_col_n) + 1

        col = 1
        for idx, code in enumerate(sorted_codes, 2):
            names = langs[code]

            try:
                # Use English name if name of the language in the current set
                # locale is unknown
                label = names[0] or names[2]

            except TypeError:
                # Happens when no name for the language codes is available.
                # "names" is "None" in that case.
                label = code
                tooltip = f'Language code "{code}" unknown.'

            else:
                # Add language name in its native representation
                # if Native letters available in current font
                # but prevent duplications like e.g. "Deutsch\nDeutsch"
                if label != names[1] and qttools.can_render(names[1], widget):
                    label = f'{names[1]}\n{label}'

                # Tooltip: Language code
                tooltip = f'{names[2]} ({code})'

                # Tooltip: completeness of translation
                try:
                    complete = languages.completeness[code]
                except KeyError:
                    pass
                else:
                    tooltip = tooltip + '\n' \
                        + _('Translated: {percent}').format(
                            percent=f'{complete}%')

            # Create button
            r = self._create_radio_button(code, label, tooltip)

            # Calculate buttons location
            row = idx - ((col - 1) * per_col_n)
            if row > per_col_n:
                row = 1
                col = col + 1

            # Add the button
            grid.addWidget(r, row, col)

        return widget

    def slot_radio(self, _):
        """Radio button state toggled."""
        btn = self.sender()

        if btn.isChecked():
            self.language_code = btn.lang_code



# SPDX-FileCopyrightText: © 2015-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.

class BackInTimeException(Exception):
    pass


class MountException(BackInTimeException):
    pass


class NoPubKeyLogin(MountException):
    pass


class KnownHost(MountException):
    pass


class HashCollision(BackInTimeException):
    pass


class EncodeValueError(BackInTimeException):
    pass


class StopException(BackInTimeException):
    pass


class Timeout(BackInTimeException):
    pass


class LastSnapshotSymlink(BackInTimeException):
    pass


class InvalidChar(BackInTimeException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class InvalidCmd(BackInTimeException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class LimitExceeded(BackInTimeException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class PermissionDeniedByPolicy(BackInTimeException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg



#!/usr/bin/env python3
# SPDX-FileCopyrightText: © 2012-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2023 Christian Buhtz <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""This script is a helper to create a manpage about Back In Times's config
file.

The file `common/config.py` is parsed for variable names, default values and
other information. The founder of that script @Germar gave a detailed
description about that script in #1354.

The script reads every line and tries to analyze it:
  - It searches for `DEFAULT` and puts those into a `dict` for later replacing
    the variable with the value.
  - If that didn't match it will look for lines starting with `#?` which is
    basically my own description for the manpage-entry.
    Multiple lines will get merged and stored in `commentline` until the
    processing of the current config option is done. That will reset
    `commentline`.
  - If a line starts with `#` it will be skipped.
  - Next the script searches for lines which ``return`` the config value (like
    `snapshots.ssh.port`. There it will extract the
    key/name (`snapshots.ssh.port`), the default value (`22`),
    the instance (`Int`) and if it is a profile or general value.
  - If the line contains a `List` value like `snapshots.include` it will
    process all values for the list like `snapshots.include.<I>.value` and
    `snapshots.include.<I>.type`.  Also it will add the size like
    `snapshots.include.size`.

In `process_line` it will replace some information with those I wrote manually
in the `#?` description, separated by `;` there is the comment,  value,
force_default and force_var. If there is no forced value it will chose the
value based on the instance with `select_values`
"""

import re
import os
import sys
from time import strftime, gmtime

PATH = os.path.join(os.getcwd(), 'common')

CONFIG = os.path.join(PATH, 'config.py')
MAN = os.path.join(PATH, 'man/C/backintime-config.1')

with open(os.path.join(PATH, '../VERSION'), 'r') as f:
    VERSION = f.read().strip('\n')

SORT = True  # True = sort by alphabet; False = sort by line numbering

c_list = re.compile(r'.*?self\.(?!set)((?:profile)?)(List)Value ?\( ?[\'"](.*?)[\'"], ?((?:\(.*\)|[^,]*)), ?[\'"]?([^\'",\)]*)[\'"]?')
c = re.compile(r'.*?self\.(?!set)((?:profile)?)(.*?)Value ?\( ?[\'"](.*?)[\'"] ?(%?[^,]*?), ?[\'"]?([^\'",\)]*)[\'"]?')

HEADER = r'''.TH backintime-config 1 "%s" "version %s" "USER COMMANDS"
.SH NAME
config \- BackInTime configuration files.
.SH SYNOPSIS
~/.config/backintime/config
.br
/etc/backintime/config
.SH DESCRIPTION
Back In Time was developed as pure GUI program and so most functions are only
usable with backintime-qt. But it is possible to use
Back In Time e.g. on a headless server. You have to create the configuration file
(~/.config/backintime/config) manually. Look inside /usr/share/doc/backintime\-common/examples/ for examples.
.PP
The configuration file has the following format:
.br
keyword=arguments
.PP
Arguments don't need to be quoted. All characters are allowed except '='.
.PP
Run 'backintime check-config' to verify the configfile, create the snapshot folder and crontab entries.
.SH POSSIBLE KEYWORDS
''' % (strftime('%b %Y', gmtime()), VERSION)

FOOTER = r'''.SH SEE ALSO
backintime, backintime-qt, backintime-askpass.
.PP
Back In Time also has a website: https://github.com/bit-team/backintime
.SH AUTHOR
This manual page was written by BIT Team(<bit-dev@python.org>).
'''

INSTANCE = 'instance'
NAME = 'name'
VALUES = 'values'
DEFAULT = 'default'
COMMENT = 'comment'
REFERENCE = 'reference'
LINE = 'line'

def groff_indented_paragraph(label: str, indent: int=6) -> str:
    """.IP - Indented Paragraph"""
    return f'.IP "{label}" {indent}'

def groff_italic(text: str) -> str:
    """\\fi - Italic"""
    return f'\\fI{text}\\fR'

def groff_indented_block(text: str) -> str:
    """
    .RS - Start indented block
    .RE - End indented block
    """
    return f'\n.RS\n{text}\n.RE\n'

def groff_linebreak() -> str:
    """.br - Line break"""
    return '.br\n'

def groff_paragraph_break() -> str:
    """.PP - Paragraph break"""
    return '.PP\n'


def output(instance='', name='', values='', default='',
           comment='', reference='', line=0):
    """Generate GNU Troff (groff) markup code for the given config entry."""
    if not default:
        default = "''"

    ret = f'Type: {instance.lower():<10}Allowed Values: {values}\n'
    ret += groff_linebreak()
    ret += f'{comment}\n'
    ret += groff_paragraph_break()

    if SORT:
        ret += f'Default: {default}'
    else:
        ret += f'Default: {default:<18} {reference} line: {line}'

    ret = groff_indented_block(ret)
    ret = groff_indented_paragraph(groff_italic(name)) + ret

    return ret


def select(a, b):
    if a:
        return a

    return b


def select_values(instance, values):
    if values:
        return values

    return {
        'bool': 'true|false',
        'str': 'text',
        'int': '0-99999'
    }[instance.lower()]


def process_line(d, key, profile, instance, name, var, default, commentline,
                 values, force_var, force_default, replace_default, counter):
    """Parsing the config.py Python code"""
    # Ignore commentlines with #?! and 'config.version'
    comment = None

    if not commentline.startswith('!') and key not in d:
        d[key] = {}
        commentline = commentline.split(';')

        try:
            comment = commentline[0]
            values = commentline[1]
            force_default = commentline[2]
            force_var = commentline[3]

        except IndexError:
            pass

        if default.startswith('self.') and default[5:] in replace_default:
            default = replace_default[default[5:]]

        if isinstance(force_default, str) \
           and force_default.startswith('self.') \
           and force_default[5:] in replace_default:
            force_default = replace_default[force_default[5:]]

        if instance.lower() == 'bool':
            default = default.lower()

        d[key][INSTANCE] = instance
        d[key][NAME] = re.sub(
            r'%[\S]', '<%s>' % select(force_var, var).upper(), name
        )
        d[key][VALUES] = select_values(instance, values)
        d[key][DEFAULT] = select(force_default, default)
        d[key][COMMENT] = re.sub(r'\\n', '\n.br\n', comment)
        d[key][REFERENCE] = 'config.py'
        d[key][LINE] = counter


def main():
    d = {
        'profiles.version': {
            INSTANCE: 'int',
            NAME: 'profiles.version',
            VALUES: '1',
            DEFAULT: '1',
            COMMENT: 'Internal version of profiles config.',
            REFERENCE: 'configfile.py',
            LINE: 419
        },
        'profiles': {
            INSTANCE: 'str',
            NAME: 'profiles',
            VALUES: 'int separated by colon (e.g. 1:3:4)',
            DEFAULT: '1',
            COMMENT: 'All active Profiles (<N> in profile<N>.snapshots...).',
            REFERENCE: 'configfile.py',
            LINE: 472
        },
        'profile<N>.name': {
            INSTANCE: 'str',
            NAME: 'profile<N>.name',
            VALUES: 'text',
            DEFAULT: 'Main profile',
            COMMENT: 'Name of this profile.',
            REFERENCE: 'configfile.py',
            LINE: 704
        }
    }

    # Default variables and there values collected from config.py
    replace_default = {}

    # Variables named "CONFIG_VERSION" or with names starting with "DEFAULT"
    regex_default = re.compile(r'(^DEFAULT[\w]*|CONFIG_VERSION)[\s]*= (.*)')

    with open(CONFIG, 'r') as f:
        print(f'Read and parse "{CONFIG}".')
        commentline = ''
        values = force_var = force_default = instance \
            = name = var = default = None

        for counter, line in enumerate(f, 1):
            line = line.lstrip()

            # parse for DEFAULT variable
            m_default = regex_default.match(line)

            # DEFAULT variable
            if m_default:
                replace_default[m_default.group(1)] \
                    = m_default.group(2)
                continue

            # Comment intended to use for the manpage
            if line.startswith('#?'):
                if commentline \
                   and ';' not in commentline \
                   and not commentline.endswith('\\n'):
                    commentline += ' '

                commentline += line.lstrip('#?').rstrip('\n')

                continue

            # Simple comments are ignored
            if line.startswith('#'):
                commentline = ''
                continue

            # e.g. "return self.profileListValue('snapshots.include', ('str:value', 'int:type'), [], profile_id)"
            # becomes
            # ('profile', 'List', 'snapshots.include', "('str:value', 'int:type')", '[]')
            m = c_list.match(line)

            if not m:
                # e.g. "return self.profileBoolValue('snapshots.use_checksum', False, profile_id)"
                # becomes
                # ('profile', 'Bool', 'snapshots.use_checksum', '', 'False')
                m = c.match(line)

            # Ignore undocumented (without "#?" comments) variables.
            if m and not commentline:
                continue

            if m:
                profile, instance, name, var, default = m.groups()

                if profile == 'profile':
                    name = 'profile<N>.' + name

                var = var.lstrip('% ')

                if instance.lower() == 'list':

                    type_key = [x.strip('"\'') for x in re.findall(r'["\'].*?["\']', var)]

                    commentline_split = commentline.split('::')

                    for i, tk in enumerate(type_key):
                        t, k = tk.split(':', maxsplit=1)

                        process_line(
                            d,
                            key,
                            profile,
                            'int',
                            '%s.size' % name,
                            var,
                            r'\-1',
                            'Quantity of %s.<I> entries.' % name,
                            values,
                            force_var,
                            force_default,
                            replace_default,
                            counter)

                        key = '%s.%s' % (name, k)
                        key = key.lower()

                        process_line(
                            d,
                            key,
                            profile,
                            t,
                            '%s.<I>.%s' % (name, k),
                            var,
                            '',
                            commentline_split[i],
                            values,
                            force_var,
                            force_default,
                            replace_default,
                            counter
                        )

                else:
                    key = re.sub(r'%[\S]', var, name).lower()

                process_line(
                    d,
                    key,
                    profile,
                    instance,
                    name,
                    var,
                    default,
                    commentline,
                    values,
                    force_var,
                    force_default,
                    replace_default,
                    counter
                )

                values = force_var = force_default = instance \
                    = name = var = default = None

                commentline = ''

    """
    Example for content of 'd':
        {
            "profiles": {
            "instance": "str",
            "name": "profiles",
            "values": "int separated by colon (e.g. 1:3:4)",
            "default": "1",
            "comment": "All active Profiles (<N> in profile<N>.snapshots...).",
            "reference": "configfile.py",
            "line": 472
        },
            "profile<N>.name": {
            "instance": "str",
            "name": "profile<N>.name",
            "values": "text",
            "default": "Main profile",
            "comment": "Name of this profile.",
            "reference": "configfile.py",
            "line": 704
        }
    """
    with open(MAN, 'w') as f:
        print(f'Write GNU Troff (groff) markup to "{MAN}". {SORT=}')
        f.write(HEADER)

        if SORT:
            # Sort by alphabet
            s = lambda x: x
        else:
            # Sort by line numbering (in the source file)
            s = lambda x: d[x][LINE]

        f.write('\n'.join(output(**d[key]) for key in sorted(d, key=s)))
        f.write(FOOTER)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import sys
import tools

tools.registerBackintimePath('common')
tools.registerBackintimePath('plugins')
tools.registerBackintimePath('common', 'plugins')
tools.registerBackintimePath('qt', 'plugins')

import logger
from exceptions import StopException


class Plugin:
    """ Interface methods to customize behavior for different backup steps

    Back In Time allows to inform plugins (implemented in Python
    files) about different steps ("events") in the backup process.
    Plugins may implement special behavior to predefined
    "events" that are declared in this interface class
    as methods.

    To implement a new plugin create a new
    class that inherits from this one and implement all
    methods.

    Plugins are loaded by calling :py:func:`PluginManager.load`.
    """

    def init(self, snapshots):
        return True

    def isGui(self):
        """Indicates a GUI-related plugin

        The return value shall indicate if the plugin
        is related to the Back In Time GUI.
        Loaded GUI-related plugins are called before non-GUI-related
        plugins by the PluginManager.

        Returns:
            True if plugin is GUI-related, otherwise False
        """
        return False

    def processBegin(self):
        """Called before a backup process is started.

        A new snapshot is only taken if required (as configured).

        Returns:
            ``None`` (return value will be ignored anyhow)
        """
        return

    def processEnd(self):
        """Called after a backup process has ended

        Returns:
            ``None`` (return value will be ignored anyhow)
        """
        return

    def error(self, code, message):
        """Indicates errors during the backup process

        Called to send errors in the backup process
        (while taking a snapshot) to plugins.

        Args:
            code: A Back In Time error code

                Known error codes:

                1: No or no valid configuration
                   (check the configuration file)
                2: A backup process is already running.
                   Make sure that automatic and manual backups
                   do not run at once.
                3: Snapshots directory not found
                   (eg. when a removable drive is not mounted)
                4: The requested snapshot for "now" already exists.
                   ``message`` contains the SID (snapshot ID) then.
                5: Error while taking a snapshot.
                   ``message`` contains more information (as string).
                6: New snapshot taken but with errors.
                   ``message`` contains the SID (snapshot ID) then.

            message: The error message for the code
                (mostly an empty string by default)

        Returns:
             ``None`` (return value will be ignored anyhow)
        """
        return

    def newSnapshot(self, snapshot_id, snapshot_path):
        """ Called when the backup process has taken a new snapshot.

        A new snapshot is only taken by the backup process
        if required (as configured).

        Args:
            snapshot_id:   The id of the new snapshot
            snapshot_path: The path to the new snapshot
        Returns:
            ``None`` (return value will be ignored anyhow)
        """
        return

    def message(self, profile_id, profile_name, level, message, timeout):
        """Send snapshot-related messages to plugins.

        Args:
            profile_id: Profile ID from configuration.
            profile_name: Profile name from the configuration.
            level: 0 = INFO, 1 = ERROR
            message: Message text.
            timeout: Requested timeout in seconds to process the message.
                Not used at the moment. (default -1 means "no timeout")
        """
        pass

    def appStart(self):
        """ Called when the GUI of Back In Time was started.

        Not called when only the CLI command was started without the GUI.

        Returns:
            ``None`` (return value will be ignored anyhow)
        """
        return

    def appExit(self):
        """ Called when the GUI of Back In Time is closed

        Returns:
            ``None`` (return value will be ignored anyhow)
        """
        return

    def mount(self, profileID = None):
        """ Called when mounting a filesystem for the profile may be necessary.

        Args:
            profileID: Profile ID from the configuration

        Returns:
            ``None`` (return value will be ignored anyhow)
        """
        return

    def unmount(self, profileID = None):
        """ Called when unmounting a filesystem for a profile may be necessary

        Args:
            profileID: Profile ID from the configuration

        Returns
            ``None`` (return value will be ignored anyhow)
        """
        return


class PluginManager:
    """ Central interface for loading plugins and calling their API

    Back In Time allows to inform plugins (implemented in Python
    files) about different steps ("events") in the backup process.

    Use this class to load installed plugin classes
    and call their methods (see the interface declared by
    :py:class:`Plugin`).

    Plugins are loaded by calling :py:func:`PluginManager.load`.

    When you call a plugin function of the PluginManager it will
    call this plugin function for all loaded plugins.
    """
    # TODO 09/28/2022: Should inherit from + implement class "Plugin"

    def __init__(self):
        self.plugins = []
        self.hasGuiPlugins = False
        self.loaded = False

    def load(self, snapshots=None, cfg=None, force=False):
        """Loads plugins

        Loads all plugins from python source code files that are stored
        in one of these plugin sub folders in the installation
        root folder:

            'plugins', 'common/plugins', 'qt/plugins'

        Plugins must inherit from :py:class:`Plugin` otherwise they
        are silently ignored.

        Args:
            snapshots (snapshots.Snapshots): Snapshot info
            cfg (config.Config): Current configuration
            force (bool): ``True`` to enforce reloading all plugins (``False``
                does only load if not already done)
        """
        if self.loaded and not force:
            return

        if snapshots is None:
            import snapshots as snapshots_
            snapshots = snapshots_.Snapshots(cfg)

        self.loaded = True
        self.plugins = []
        self.hasGuiPlugins = False

        loadedPlugins = []

        # TODO 09/28/2022: Move hard coded plugin folders to configuration
        for path in ('plugins', 'common/plugins', 'qt/plugins'):
            fullPath = tools.backintimePath(path)

            if os.path.isdir(fullPath):
                logger.debug('Register plugin path %s' %fullPath, self)
                tools.registerBackintimePath(path)

                for f in os.listdir(fullPath):

                    if f not in loadedPlugins and f.endswith('.py') and not f.startswith('__'):
                        # logger.debug('Probing plugin %s' % f, self)

                        try:
                            module = __import__(f[: -3])
                            module_dict = module.__dict__

                            for key, value in list(module_dict.items()):
                                if key.startswith('__'):
                                    continue

                                if type(value) is type:
                                    # A plugin must implement this class via inheritance
                                    if issubclass(value, Plugin):
                                        plugin = value()
                                        if plugin.init(snapshots):
                                            logger.debug('Add plugin %s' %f, self)
                                            if plugin.isGui():
                                                self.hasGuiPlugins = True
                                                self.plugins.insert(0, plugin)
                                            else:
                                                self.plugins.append(plugin)
                            loadedPlugins.append(f)

                        except BaseException as e:
                            logger.error('Failed to load plugin %s: %s' %(f, str(e)), self)

    def processBegin(self):
        ret_val = True
        for plugin in self.plugins:
            try:
                plugin.processBegin()
            except StopException:
                ret_val = False
            except BaseException as e:
                self.logError(plugin, e)
        return ret_val

    def processEnd(self):
        for plugin in reversed(self.plugins):
            try:
                plugin.processEnd()
            except BaseException as e:
                self.logError(plugin, e)

    def error(self, code, message = ''):
        for plugin in self.plugins:
            try:
                plugin.error(code, message)
            except BaseException as e:
                self.logError(plugin, e)

    def newSnapshot(self, snapshot_id, snapshot_path):
        for plugin in self.plugins:
            try:
                plugin.newSnapshot(snapshot_id, snapshot_path)
            except BaseException as e:
                self.logError(plugin, e)

    def message(self, profile_id, profile_name, level, message, timeout = -1):
        for plugin in self.plugins:
            try:
                plugin.message(profile_id, profile_name, level, message, timeout)
            except BaseException as e:
                self.logError(plugin, e)

    def appStart(self):
        for plugin in reversed(self.plugins):
            try:
                plugin.appStart()
            except BaseException as e:
                self.logError(plugin, e)

    def appExit(self):
        for plugin in reversed(self.plugins):
            try:
                plugin.appExit()
            except BaseException as e:
                self.logError(plugin, e)

    def mount(self, profileID = None):
        for plugin in reversed(self.plugins):
            try:
                plugin.mount(profileID)
            except BaseException as e:
                self.logError(plugin, e)

    def unmount(self, profileID = None):
        for plugin in reversed(self.plugins):
            try:
                plugin.unmount(profileID)
            except BaseException as e:
                self.logError(plugin, e)

    def logError(self, plugin, e):
        logger.error('Plugin %s %s failed: %s'
                     %(plugin.__module__,               #plugin name
                       sys._getframe(1).f_code.co_name, #method name
                       str(e)),                         #exception
                     self, 1)
